from contextlib import contextmanager
import logging
import os
import timeit
import time

from mdt import get_processing_strategy
from mdt.utils import load_samples, per_model_logging_context
from mdt.lib.processing_strategies import SamplingProcessor, SaveAllSamples, \
    SaveNoSamples, get_full_tmp_results_path, SaveSpecificMaps
from mdt.lib.exceptions import InsufficientProtocolError


__author__ = 'Robbert Harms'
__date__ = "2015-05-01"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def sample_composite_model(model, input_data, output_folder, nmr_samples, thinning, burnin, tmp_dir,
                           method=None, recalculate=False, store_samples=True, sample_items_to_save=None,
                           initialization_data=None):
    """Sample a composite model.

    Args:
        model (:class:`~mdt.models.composite.DMRICompositeModel`): a composite model to sample
        input_data (:class:`~mdt.utils.MRIInputData`): The input data object with which the model
            is initialized before running
        output_folder (string): The full path to the folder where to place the output
        nmr_samples (int): the number of samples we would like to return.
        burnin (int): the number of samples to burn-in, that is, to discard before returning the desired
            number of samples
        thinning (int): how many sample we wait before storing a new one. This will draw extra samples such that
                the total number of samples generated is ``nmr_samples * (thinning)`` and the number of samples stored
                is ``nmr_samples``. If set to one or lower we store every sample after the burn in.
        tmp_dir (str): the preferred temporary storage dir
        method (str): The sampling method to use, one of:
            - 'AMWG', for the Adaptive Metropolis-Within-Gibbs method
            - 'SCAM', for the Single Component Adaptive Metropolis
            - 'FSL', for the sampling method used in the FSL toolbox
            - 'MWG', for the Metropolis-Within-Gibbs (simple random walk metropolis without updates)

            If not given, defaults to 'AMWG'.
        recalculate (boolean): If we want to recalculate the results if they are already present.
        store_samples (boolean, sequence or :class:`mdt.lib.processing_strategies.SamplesStorageStrategy`): if set to False
            we will store none of the samples. If set to True we will save all samples. If set to a sequence we expect a
            sequence of integer numbers with sample positions to store. Finally, you can also give a subclass instance
            of :class:`~mdt.lib.processing_strategies.SamplesStorageStrategy` (it is then typically set to
            a :class:`mdt.lib.processing_strategies.SaveThinnedSamples` instance).
        sample_items_to_save (list): list of output names we want to store the samples of. If given, we only
            store the items specified in this list. Valid items are the free parameter names of the model and the
            items 'LogLikelihood' and 'LogPrior'.
        initialization_data (:class:`~mdt.utils.InitializationData`): provides (extra) initialization data to use
            during model fitting. If we are optimizing a cascade model this data only applies to the last model in the
            cascade.
    """
    samples_storage_strategy = SaveAllSamples()
    if store_samples:
        if sample_items_to_save:
            samples_storage_strategy = SaveSpecificMaps(included=sample_items_to_save)
    else:
        samples_storage_strategy = SaveNoSamples()

    if not model.is_input_data_sufficient(input_data):
        raise InsufficientProtocolError(
            'The provided protocol is insufficient for this model. '
            'The reported errors where: {}'.format(model.get_input_data_problems(input_data)))

    logger = logging.getLogger(__name__)

    if not recalculate:
        if os.path.exists(os.path.join(output_folder, 'UsedMask.nii.gz')) \
                or os.path.exists(os.path.join(output_folder, 'UsedMask.nii')):
            logger.info('Not recalculating {} model'.format(model.name))
            return load_samples(output_folder)

    if not os.path.isdir(output_folder):
        os.makedirs(output_folder)

    model.set_input_data(input_data)

    with per_model_logging_context(output_folder, overwrite=recalculate):
        if initialization_data:
            logger.info('Preparing the model with the user provided initialization data.')
            initialization_data.apply_to_model(model, input_data)

        with _log_info(logger, model.name):
            worker = SamplingProcessor(
                nmr_samples, thinning, burnin, method or 'AMWG',
                model, input_data.mask, input_data.nifti_header, output_folder,
                get_full_tmp_results_path(output_folder, tmp_dir), recalculate,
                samples_storage_strategy=samples_storage_strategy)

            processing_strategy = get_processing_strategy('sampling')
            return processing_strategy.process(worker)


@contextmanager
def _log_info(logger, model_name):
    minimize_start_time = timeit.default_timer()
    logger.info('Sampling {} model'.format(model_name))
    yield
    run_time = timeit.default_timer() - minimize_start_time
    run_time_str = time.strftime('%H:%M:%S', time.gmtime(run_time))
    logger.info('Sampled {0} model with runtime {1} (h:m:s).'.format(model_name, run_time_str))



import inspect
import re
from copy import deepcopy
import numpy as np
import tatsu

from mdt.component_templates.base import ComponentBuilder, ComponentTemplate
from mdt.lib.components import get_component
from mdt.models.composite import DMRICompositeModel
from mot.lib.cl_function import CLFunction, SimpleCLFunction
from mdt.model_building.trees import CompartmentModelTree
import collections

__author__ = 'Robbert Harms'
__date__ = "2017-02-14"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


_composite_model_expression_parser = tatsu.compile('''
    result = expr;
    expr = term ('+'|'-') expr | term;
    term = factor ('*'|'/') term | factor;
    factor = '(' expr ')' | model;
    model = model_name ['(' nickname ')'];
    model_name = /[a-zA-Z_]\w*/;
    nickname = /[a-zA-Z_]\w*/;
''')


class DMRICompositeModelBuilder(ComponentBuilder):

    def _create_class(self, template):
        """Creates classes with as base class DMRICompositeModel

        Args:
            template (CompositeModelTemplate): the composite model config template
                to use for creating the class with the right init settings.
        """
        class AutoCreatedDMRICompositeModel(DMRICompositeModel):

            def __init__(self, model_name=None, **kwargs):
                model_name = model_name or deepcopy(template.name)

                keywords = {'signal_noise_model': deepcopy(template.signal_noise_model),
                            'enforce_weights_sum_to_one': template.enforce_weights_sum_to_one}
                keywords.update(kwargs)

                super().__init__(
                    model_name,
                    CompartmentModelTree(parse_composite_model_expression(template.model_expression)),
                    deepcopy(_resolve_likelihood_function(template.likelihood_function)),
                    **keywords
                )

                for full_param_name, value in template.inits.items():
                    self.init(full_param_name, deepcopy(value))

                for full_param_name, value in template.fixes.items():
                    self.fix(full_param_name, deepcopy(value))

                for full_param_name, value in template.lower_bounds.items():
                    self.set_lower_bound(full_param_name, deepcopy(value))

                for full_param_name, value in template.upper_bounds.items():
                    self.set_upper_bound(full_param_name, deepcopy(value))

                self.nmr_parameters_for_bic_calculation = self.get_nmr_parameters()

                if template.sort_maps:
                    self._post_optimization_modifiers.append(_get_map_sorting_modifier(
                        template.sort_maps, self._model_functions_info.get_model_parameter_list()))

                self._post_optimization_modifiers.extend(_get_model_post_optimization_modifiers(
                    self._model_functions_info.get_model_list()))
                self._post_optimization_modifiers.extend(deepcopy(template.post_optimization_modifiers))

                self._extra_optimization_maps_funcs.extend(_get_model_extra_optimization_maps_funcs(
                    self._model_functions_info.get_model_list()))
                self._extra_optimization_maps_funcs.extend(deepcopy(template.extra_optimization_maps))

                self._extra_sampling_maps_funcs.extend(_get_model_extra_sampling_maps_funcs(
                    self._model_functions_info.get_model_list()))
                self._extra_sampling_maps_funcs.extend(deepcopy(template.extra_sampling_maps))

                self._proposal_callbacks.extend(_get_model_proposal_callbacks(
                    self._model_functions_info.get_model_list()))

                self._model_priors.extend(_resolve_model_prior(
                    template.extra_prior, self._model_functions_info.get_model_parameter_list()))

            def _get_suitable_volume_indices(self, input_data):
                volume_selection = template.volume_selection

                if not volume_selection:
                    return super()._get_suitable_volume_indices(input_data)

                use_unweighted = volume_selection.get('use_unweighted', True)
                use_weighted = volume_selection.get('use_weighted', True)
                unweighted_threshold = volume_selection.get('unweighted_threshold', 25e6)

                protocol = input_data.protocol

                if protocol.has_column('g') and protocol.has_column('b'):
                    if use_weighted:
                        if 'min_bval' in volume_selection and 'max_bval' in volume_selection:
                            protocol_indices = protocol.get_indices_bval_in_range(start=volume_selection['min_bval'],
                                                                                  end=volume_selection['max_bval'])
                        else:
                            protocol_indices = protocol.get_weighted_indices(unweighted_threshold)
                    else:
                        protocol_indices = []

                    if use_unweighted:
                        protocol_indices = list(protocol_indices) + \
                                           list(protocol.get_unweighted_indices(unweighted_threshold))
                else:
                    return list(range(protocol.length))

                return np.unique(protocol_indices)

        for name, method in template.bound_methods.items():
            setattr(AutoCreatedDMRICompositeModel, name, method)

        return AutoCreatedDMRICompositeModel


class CompositeModelTemplate(ComponentTemplate):
    """The cascade config to inherit from.

    These configs are loaded on the fly by the DMRICompositeModelBuilder

    Attributes:
        name (str): the name of the model, defaults to the class name

        description (str): model description

        post_optimization_modifiers (list): a list of modification callbacks to change the estimated
            optimization points. This should not add new maps, for that use the ``additional_results`` directive.
            This can be used to, for example, sort maps into a similar valued but sorted representation
            (actually the ``sort_maps`` directive creates a post_optimization_modifier)

            Examples:

            .. code-block:: python

                post_optimization_modifiers = [lambda d: reorient_tensor(d),
                                               lambda d: {'w_stick0.w': ..., 'w_stick1.w': ...}
                                                ...]

            This should return a dictionary with updated maps.

        sort_maps (list of tuple): The maps to sort voxel-by voxel as post-processing.
            To ensure that the point estimate from optimization can directly be used in MCMC sample, we sometimes
            need to rearrange the weights (and therefore the corresponding compartments). For example, some composite
            models have a prior on the weights of similar compartments to restrict them to a decreasing order, which
            is typically done to prevent bimodal continuous_distributions. This model directive allows you to easily specify
            which parameters to rearrange as first post-processing of optimization results.

            The first tuple needs to be a parameter reference or a  ``Weight`` compartment, the other tuples can
            contain either parameters or compartments.

            Example input::

                sort_maps = [('w0.w', 'w1.w'), ('Stick0', 'Stick1')]

            this input sorts the weights w0.w and w1.w and afterwards sorts all the parameters of the Stick0 and Stick1
            compartment according to the sorting of those weights.

            One can also sort just one parameter of a compartment using::

                sort_maps = [('w0', 'w1'), ('Stick0.theta', 'Stick1.theta')]

            which will again sort the weights but will only apply the sorting on the theta map of both the Stick
            compartments.

        extra_optimization_maps (list): a list of functions to return extra information maps based on a point estimate.
            This is called after the post optimization modifiers and after the model calculated uncertainties based
            on the Fisher Information Matrix. Therefore, these routines can propagate uncertainties in the estimates.
            Please note that this is only for adding additional maps. For changing the point estimate of the
            optimization, please use the ``post_optimization_modifiers`` directive.

            Examples::

                extra_optimization_maps = [lambda d: {'FS': 1 - d['w_ball.w']},
                                           lambda d, protocol: {'Kurtosis.MK': <...>},
                                           lambda d: {'Power2': d['foo']**2, 'Power3': d['foo']**3},
                                           ...]

        extra_sampling_maps (list): a list of functions to return additional maps as results from sample.
            This is called after sample with as argument a dictionary containing the sample results and
            the values of the fixed parameters.

            Examples::

                extra_sampling_maps = [lambda d: {'FS': np.mean(d['w_stick0.w'], axis=1),
                                                  'FS.std': np.std(d['w_stick0.w'], axis=1)}
                                      ...]

        model_expression (str): the model expression. For the syntax see:
            mdt.models.parsers.CompositeModelExpression.ebnf

        likelihood_function (:class:`mdt.model_building.likelihood_functions.LikelihoodFunction` or str): the
            likelihood function to use during optimization, can also can be a string with one of
            'Gaussian', 'OffsetGaussian' or 'Rician'

        signal_noise_model (SignalNoiseModel): optional signal noise decorator

        inits (dict): indicating the initialization values for the parameters. Example:

            .. code-block:: python

                inits = {'Stick.theta': np.pi}

        fixes (dict): indicating the constant value for the given parameters. Example:

            .. code-block:: python

                fixes = {'Ball.d': 3.0e-9,
                         'NODDI_EC.kappa': SimpleAssignment('NODDI_IC.kappa'),
                         'NODDI_EC.theta': 'NODDI_IC.theta'}

            Next to values, this also accepts strings as dependencies (or dependecy objects directly).

        upper_bounds (dict): indicating the upper bounds for the given parameters. Example:

            .. code-block:: python

                upper_bounds = {'Stick.theta': pi}

        lower_bounds (dict): indicating the lower bounds for the given parameters. Example:

            .. code-block:: python

                lower_bounds = {'Stick.theta': 0}

        enforce_weights_sum_to_one (boolean): set to False to disable the automatic Weight-sum-to-one dependency.
            By default it is True and we add them.

        volume_selection (dict): the volume selection by this model. This can be used to limit the volumes used
            in the analysis to only the volumes included in the specification.
            Set to None, or an empty dict to disable.
            The options available are:

               * ``unweighted_threshold`` (float): the threshold differentiating between
                 weighted and unweighted volumes
               * ``use_unweighted`` (bool): if we want to use unweighted volumes or not
               * ``use_weighted`` (bool): if we want to use the diffusion weigthed volumes or not
               * ``min_bval`` (float): the minimum b-value to include
               * ``max_bval`` (float): the maximum b-value to include

            If the method ``_get_suitable_volume_indices`` is overwritten, this does nothing.

        prior (str, mdt.model_building.utils.ModelPrior or None): a model wide prior. This is used in conjunction with
            the compartment priors and the parameter priors. If a string is given we will automatically construct a
            :class:`mdt.model_building.utils.ModelPrior` from that string.
    """
    _component_type = 'composite_models'
    _builder = DMRICompositeModelBuilder()

    name = ''
    description = ''
    post_optimization_modifiers = []
    extra_optimization_maps = []
    extra_sampling_maps = []
    model_expression = ''
    likelihood_function = 'OffsetGaussian'
    signal_noise_model = None
    inits = {}
    fixes = {}
    upper_bounds = {}
    lower_bounds = {}
    enforce_weights_sum_to_one = True
    volume_selection = None
    extra_prior = None
    sort_maps = None

    @classmethod
    def meta_info(cls):
        meta_info = deepcopy(ComponentTemplate.meta_info())
        meta_info.update({'name': cls.name,
                          'description': cls.description})
        return meta_info


def _resolve_likelihood_function(likelihood_function):
    """Resolve the likelihood function from string if necessary.

    The composite models accept likelihood functions as a string and as a object. This function
    resolves the strings if a string is given, else it returns the object passed.

    Args:
        likelihood_function (str or object): the likelihood function to resolve to an object

    Returns:
        mdt.model_building.likelihood_models.LikelihoodFunction: the likelihood function to use
    """
    if isinstance(likelihood_function, str):
        return get_component('likelihood_functions', likelihood_function + 'LikelihoodFunction')()
    else:
        return likelihood_function


def _resolve_model_prior(prior, model_parameters):
    """Resolve the model priors.

    Args:
        prior (None or str or mot.lib.cl_function.CLFunction): the prior defined in the composite model template.
        model_parameters (str): the (model, parameter) tuple for all the parameters in the model

    Returns:
        list of mdt.model_building.utils.ModelPrior: list of model priors
    """
    if prior is None:
        return []

    if isinstance(prior, CLFunction):
        return [prior]

    parameters = []
    for m, p in model_parameters:
        dotted_name = '{}.{}'.format(m.name, p.name)
        bar_name = dotted_name.replace('.', '_')

        if dotted_name in prior:
            prior = prior.replace(dotted_name, bar_name)
            parameters.append(('mot_float_type', dotted_name))
        elif bar_name in prior:
            parameters.append(('mot_float_type', dotted_name))

    return [SimpleCLFunction('mot_float_type', 'model_prior', parameters, prior)]


def _get_map_sorting_modifier(sort_maps, model_list):
    """Construct the map sorting modification routine for the given maps_to_sort attribute.

    Args:
        sort_maps (list of tuple): the list with compartments/parameters to sort. Sorting is done on the first element and
            the other maps are sorted based on that indexing.
        model_list (list of tuple): the list of model, parameter tuples to use for matching names.
    Returns:
        tuple: name, modification routine. The modification routine will sort the maps as specified.
    """
    def is_param_name(name):
        return '.' in name

    def get_model_by_name(name):
        for model, parameter in model_list:
            if model.name == name:
                return model

    sort_keys = []
    for name in sort_maps[0]:
        for model, parameter in model_list:
            full_name = '{}.{}'.format(model.name, parameter.name)
            if model.name == name or full_name == name:
                sort_keys.append(full_name)
                break

    other_sort_pairs = []
    for other_pair in sort_maps[1:]:
        if is_param_name(other_pair[0]):
            other_sort_pairs.append(other_pair)
        else:
            model = get_model_by_name(other_pair[0])
            for parameter in model.get_free_parameters():
                sort_pair = []
                for model_name in other_pair:
                    sort_pair.append('{}.{}'.format(model_name, parameter.name))
                other_sort_pairs.append(sort_pair)

    names = []
    names.extend(sort_keys)
    for pairs in other_sort_pairs:
        names.extend(pairs)

    def map_sorting(results):
        sort_matrix = np.column_stack([results[map_name] for map_name in sort_keys])
        ranking = np.atleast_2d(np.squeeze(np.argsort(sort_matrix, axis=1)[:, ::-1]))
        list_index = np.arange(ranking.shape[0])

        sorted_maps = []
        sorted_maps.extend(sort_matrix[list_index, ranking[:, ind], None] for ind in range(ranking.shape[1]))

        for pair in other_sort_pairs:
            sort_matrix = np.column_stack([results[map_name] for map_name in pair])
            sorted_maps.extend(sort_matrix[list_index, ranking[:, ind], None] for ind in range(ranking.shape[1]))

        return dict(zip(names, sorted_maps))

    return map_sorting


def _get_model_post_optimization_modifiers(compartments):
    """Get a list of all the post optimization modifiers defined in the models.

    This function will add a wrapper around the modification routines to make the input and output maps relative to the
    model. That it, these functions expect the parameter names without the model name and output map names without
    the model name, whereas the expected input and output of the modifiers of the model is with the full model.map name.

    Args:
        compartments (list): the list of compartment models from which to get the modifiers

    Returns:
        list: the list of modification routines taken from the compartment models.
    """
    modifiers = []

    def get_wrapped_modifier(compartment_name, original_modifier):
        def get_compartment_specific_maps(results):
            return {k[len(compartment_name) + 1:]: v for k, v in results.items() if k.startswith(compartment_name)}

        def prepend_compartment_name(results):
            return {'{}.{}'.format(compartment_name, key): value for key, value in results.items()}

        def wrapped_modifier(results):
            return prepend_compartment_name(original_modifier(get_compartment_specific_maps(results)))

        return wrapped_modifier

    for compartment in compartments:
        if hasattr(compartment, 'post_optimization_modifiers'):
            for modifier in compartment.post_optimization_modifiers:
                modifiers.append(get_wrapped_modifier(compartment.name, modifier))

    return modifiers


def _get_model_extra_optimization_maps_funcs(compartments):
    """Get a list of all the additional result functions defined in the compartments.

    This function will add a wrapper around the modification routines to make the input and output maps relative to the
    model. That it, the functions in the compartments expect the parameter names without the model name and they output
    maps without the model name, whereas the expected input and output of the modifiers of the model is with the
    full model.map name.

    Args:
        compartments (list): the list of compartment models from which to get the modifiers

    Returns:
        list: the list of modification routines taken from the compartment models.
    """
    funcs = []

    def get_wrapped_func(compartment_name, original_func):
        def get_compartment_specific_maps(results):
            maps = {k[len(compartment_name) + 1:]: v for k, v in results.items() if k.startswith(compartment_name)}

            if 'covariances' in results and results['covariances'] is not None:
                p = re.compile(compartment_name + r'\.\w+_to_' + compartment_name + r'\.\w+')
                maps['covariances'] = {k.replace(compartment_name + '.', ''): v
                                       for k, v in results['covariances'].items() if p.match(k)}
            return maps

        def prepend_compartment_name(results):
            return {'{}.{}'.format(compartment_name, key): value for key, value in results.items()}

        if 'input_data' in inspect.signature(original_func).parameters:
            def wrapped_modifier(results, input_data):
                return prepend_compartment_name(original_func(get_compartment_specific_maps(results), input_data))
        else:
            def wrapped_modifier(results):
                return prepend_compartment_name(original_func(get_compartment_specific_maps(results)))

        return wrapped_modifier

    for compartment in compartments:
        if hasattr(compartment, 'extra_optimization_maps_funcs'):
            for func in compartment.extra_optimization_maps_funcs:
                funcs.append(get_wrapped_func(compartment.name, func))

    return funcs


def _get_model_extra_sampling_maps_funcs(compartments):
    """Get a list of all the additional post-sample functions defined in the compartments.

    This function will add a wrapper around the modification routines to make the input and output maps relative to the
    model. That it, the functions in the compartments expect the parameter names without the model name and they output
    maps without the model name, whereas the expected input and output of the modifiers of the model is with the
    full model.map name.

    Args:
        compartments (list): the list of compartment models from which to get the modifiers

    Returns:
        list: the list of extra sample routines taken from the compartment models.
    """
    funcs = []

    def get_wrapped_func(compartment_name, original_func):
        def prepend_compartment_name(results):
            return {'{}.{}'.format(compartment_name, key): value for key, value in results.items()}

        def wrapped_modifier(results):
            return prepend_compartment_name(original_func(CompartmentContextResults(compartment_name, results)))

        return wrapped_modifier

    for compartment in compartments:
        if hasattr(compartment, 'extra_sampling_maps_funcs'):
            for func in compartment.extra_sampling_maps_funcs:
                funcs.append(get_wrapped_func(compartment.name, func))

    return funcs


def _get_model_proposal_callbacks(compartments):
    """Get a list of all the additional proposal callback functions defined in the compartments.

    This function will add a wrapper around the proposal callback to make the input and output maps relative to the
    model.

    Args:
        compartments (list): the list of compartment models from which to get the modifiers

    Returns:
        list: the list of proposal callbacks taken from the compartment models.
    """
    funcs = []
    for compartment in compartments:
        if hasattr(compartment, 'proposal_callbacks'):
            for params, func in compartment.proposal_callbacks:
                compartment_params = [(compartment, p) for p in params]
                funcs.append((compartment_params, func))
    return funcs


class CompartmentContextResults(collections.Mapping):

    def __init__(self, compartment_name, input_results):
        """Translates the original results to the context of a single compartment.

        This basically adds a wrapper around the input dictionary to make the keys relative to the compartment.

        Args:
            compartment_name (str): the name of the compartment we are making things relative for
            input_results (dict): the original input we want to make relative
        """
        self._compartment_name = compartment_name
        self._input_results = input_results
        self._valid_keys = [key for key in self._input_results if key.startswith(self._compartment_name + '.')]

    def __getitem__(self, key):
        return self._input_results['{}.{}'.format(self._compartment_name, key)]

    def __len__(self):
        return len(self._valid_keys)

    def __iter__(self):
        return self._valid_keys


def parse_composite_model_expression(model_expression):
    """Parse the given model expression into a suitable model tree.

    Args:
        model_expression (str): the model expression string. Example:

        .. code-block:: none

            S0 * ( (Weight(Wball) * Ball) +
                   (Weight(Wstick) * Stick ) )

        If the model name is followed by parenthesis the string in parenthesis will represent the model's nickname.

    Returns:
        :class:`list`: the compartment model tree for use in composite models.
    """
    class Semantics(object):

        def expr(self, ast):
            if not isinstance(ast, list):
                return ast
            if isinstance(ast, list):
                return ast[0], ast[2], ast[1]
            return ast

        def term(self, ast):
            if not isinstance(ast, list):
                return ast
            if isinstance(ast, list):
                return ast[0], ast[2], ast[1]
            return ast

        def factor(self, ast):
            if isinstance(ast, list):
                return ast[1]
            return ast

        def model(self, ast):
            if isinstance(ast, str):
                return get_component('compartment_models', ast)()
            else:
                return get_component('compartment_models', ast[0])(ast[2])

    return _composite_model_expression_parser.parse(model_expression, semantics=Semantics())



from mdt import LibraryFunctionTemplate


class RotateOrthogonalVector(LibraryFunctionTemplate):

    description = '''
        Uses Rodrigues' rotation formula to rotate the given vector v by psi around k.

        This function assumes that the given vectors v and k are orthogonal which allows a speed up/

        Args:
            basis: the unit vector defining the rotation axis (k)
            to_rotate: the vector to rotate by the angle psi (v)
            psi: the rotation angle (psi)

        Returns:
            vector: the rotated vector
    '''
    return_type = 'mot_float_type4'
    parameters = [('mot_float_type4', 'basis'),
                  ('mot_float_type4', 'to_rotate'),
                  ('mot_float_type', 'psi')]
    cl_code = '''
        mot_float_type cos_psi;
        mot_float_type sin_psi = sincos(psi, &cos_psi);

        return to_rotate * cos_psi + (cross(basis, to_rotate) * sin_psi);
    '''


class RotateVector(LibraryFunctionTemplate):

    description = '''
        Uses Rodrigues' rotation formula to rotate the given vector v by psi around k.

        Args:
            basis: the unit vector defining the rotation axis (k)
            to_rotate: the vector to rotate by the angle psi (v)
            psi: the rotation angle (psi)

        Returns:
            vector: the rotated vector
    '''
    return_type = 'mot_float_type4'
    parameters = [('mot_float_type4', 'basis'),
                  ('mot_float_type4', 'to_rotate'),
                  ('mot_float_type', 'psi')]
    cl_code = '''
        mot_float_type cos_psi;
        mot_float_type sin_psi = sincos(psi, &cos_psi);

        return to_rotate * cos_psi
                + (cross(basis, to_rotate) * sin_psi)
                + (basis * dot(basis, to_rotate) * (1 - cos_psi));
    '''


class SphericalToCartesian(LibraryFunctionTemplate):

    description = '''
        """Convert polar coordinates in 3d space to cartesian unit coordinates.

        .. code-block:: python

            x = sin(theta) * cos(phi)
            y = sin(theta) * sin(phi)
            z = cos(theta)

        Args:
            theta: polar angle of the first vector
            phi: azimuth angle of the first vector
    '''
    return_type = 'mot_float_type4'
    parameters = ['theta', 'phi']
    cl_code = '''
        mot_float_type cos_theta;
        mot_float_type sin_theta = sincos(theta, &cos_theta);
        mot_float_type cos_phi;
        mot_float_type sin_phi = sincos(phi, &cos_phi);
        
        return (mot_float_type4)(cos_phi * sin_theta, sin_phi * sin_theta, cos_theta, 0.0);
    '''


class TensorSphericalToCartesian(LibraryFunctionTemplate):

    description = '''
        Generates the D matrix for a Tensor compartment.

        The angles ``theta`` and ``phi`` are used for creating the first vector, ``vec0``.
        Next, ``vec0`` is rotated 90 degrees in the (x, z) plane to form a vector perpendicular to the
        principal direction. This vector is then rotated around ``psi`` to generate the first perpendicular
        orientation, ``vec1``. The third vector is generated by being perpendicular to the other two vectors.

        Args:
            theta: polar angle of the first vector
            phi: azimuth angle of the first vector
            psi: rotation around the first vector, used to generate the perpendicular vectors.
    '''
    dependencies = ['RotateOrthogonalVector', 'SphericalToCartesian']
    parameters = ['theta', 'phi', 'psi',
                  ('mot_float_type4*', 'vec0'),
                  ('mot_float_type4*', 'vec1'),
                  ('mot_float_type4*', 'vec2')]
    cl_code = '''
        *vec0 = SphericalToCartesian(theta, phi);
        *vec1 = RotateOrthogonalVector(*vec0, SphericalToCartesian(theta + M_PI_2_F, phi), psi);
        *vec2 = cross(*vec0, *vec1);
    '''


class CartesianPolarDotProduct(LibraryFunctionTemplate):

    description = '''
        Calculates the dot product between a cartesian and a polar coordinate vector.
        
        Prior to taking the dot product it will convert the polar coordinate vector to cartesian coordinates.
         
        Args:
            v0_x: the x coordinate of the first vector
            v0_y: the y coordinate of the first vector
            v0_z: the z coordinate of the first vector
            v1_theta: the polar angle of the second vector
            v1_phi: the azimuth angle of the second vector
            
        Returns:
            dot product between the two vectors
    '''
    return_type = 'mot_float_type'
    parameters = [('mot_float_type', 'v0_x'),
                  ('mot_float_type', 'v0_y'),
                  ('mot_float_type', 'v0_z'),
                  ('mot_float_type', 'v1_theta'),
                  ('mot_float_type', 'v1_phi')]
    cl_code = '''
        mot_float_type cos_theta;
        mot_float_type sin_theta = sincos(v1_theta, &cos_theta);
        mot_float_type cos_phi;
        mot_float_type sin_phi = sincos(v1_phi, &cos_phi);
        
        return (v0_x * (cos_phi * sin_theta)) + (v0_y * (sin_phi * sin_theta)) + (v0_z * cos_theta);
    '''


class EigenvaluesSymmetric3x3(LibraryFunctionTemplate):
    description = '''
        Calculate the eigenvalues of a symmetric 3x3 matrix. 

        This simple algorithm only works in case of a real and symmetric matrix. 

        This returns the eigenvalues such that eig1 >= eig2 >= eig3, i.e. from large to small.

        References:
            [1]: https://en.wikipedia.org/wiki/Eigenvalue_algorithm#3.C3.973_matrices
            [2]: Smith, Oliver K. (April 1961), "Eigenvalues of a symmetric 3 × 3 matrix.", Communications of the ACM, 
                 4 (4): 168, doi:10.1145/355578.366316

        Args:
            A: the matrix as an array in c/row-major order
            v: the output eigenvalues as a vector of three elements.            
    '''
    parameters = [('mot_float_type*', 'A'),
                  ('mot_float_type*', 'v')]
    cl_code = '''
        double p1 = (A[1] * A[1]) + (A[2] * A[2]) + (A[5] * A[5]);

        if (p1 == 0.0){
            v[0] = A[0];
            v[1] = A[4];
            v[2] = A[8];
            return;
        }
        
        double q = (A[0] + A[4] + A[8]) / 3.0;
        double p = sqrt(((A[0] - q)*(A[0] - q) + (A[4] - q)*(A[4] - q) + (A[8] - q)*(A[8] - q) + 2*p1) / 6.0);

        double r = (
             ((A[0] - q)/p) * ((((A[4] - q)/p) * ((A[8] - q)/p)) - ((A[7]/p) * (A[5]/p))) 
            - (A[1]/p)      * ((A[3]/p) * ((A[8] - q)/p) - (A[6]/p) * (A[5]/p)) 
            + (A[2]/p)      * ((A[3]/p) * (A[7]/p) - (A[6]/p) * ((A[4] - q)/p))
        ) / 2.0;

        double phi;
        if(r <= -1){
            phi = M_PI / 3.0;
        }
        else if(r >= 1){
            phi = 0;
        }
        else{
            phi = acos(r) / 3;
        }

        v[0] = q + 2 * p * cos(phi);
        v[2] = q + 2 * p * cos(phi + (2*M_PI/3.0));
        v[1] = 3 * q - v[0] - v[2];
    '''


class SquareMatrixMultiplication(LibraryFunctionTemplate):
    description = '''
        Matrix multiplication of two square matrices. 
        
        Having this as a special function is slightly faster than a more generic matrix multiplication algorithm.
        
        All matrices are expected to be in c/row-major order.
        
        Args:
            n: the rectangular size of the matrix (the number of rows / the number of columns). 
            A: the left matrix 
            B: the right matrix 
            C: the output matrix            
    '''
    parameters = [('uint', 'n'),
                  ('mot_float_type*', 'A'),
                  ('mot_float_type*', 'B'),
                  ('mot_float_type*', 'C')]
    cl_code = '''
        int i, j, z;
        double sum;
        for(int ind = 0; ind < n*n; ind++){
            i = ind / n;
            j = ind - n * i;
            
            printf("%i, %i, %i", ind, i, j);
            
            sum = 0;
            for(z = 0; z < n; z++){
                sum += A[i * n + z] * B[z * n + j];
            }
            C[ind] = sum;
        }
    '''




#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""Launches the MDT Graphical User Interface."""
import argparse
import os
import textwrap

from argcomplete.completers import FilesCompleter

from mdt.gui.model_fit.qt_main import start_gui
from mdt.lib.shell_utils import BasicShellApplication

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class GUI(BasicShellApplication):

    def _get_arg_parser(self, doc_parser=False):
        description = textwrap.dedent(__doc__)

        parser = argparse.ArgumentParser(description=description, formatter_class=argparse.RawTextHelpFormatter)
        parser.add_argument('-d', '--dir', metavar='dir', type=str, help='the base directory for the file choosers',
                            default=None).completer = FilesCompleter()
        return parser

    def run(self, args, extra_args):
        if args.dir:
            cwd = os.path.realpath(args.dir)
        else:
            cwd = os.getcwd()

        start_gui(cwd)


def get_doc_arg_parser():
    return GUI().get_documentation_arg_parser()


if __name__ == '__main__':
    GUI().start()



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""Create a (brain) mask for the given DWI. This uses the median-otsu algorithm."""
import argparse
import logging
import os
import mdt
from argcomplete.completers import FilesCompleter

import mot.configuration
from mdt.lib.shell_utils import BasicShellApplication
from mot.lib import cl_environments
from mot.lib.load_balance_strategies import EvenDistribution
import textwrap

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CreateMask(BasicShellApplication):

    def __init__(self):
        super().__init__()
        self.available_devices = {ind: env for ind, env in
                                  enumerate(cl_environments.CLEnvironmentFactory.smart_device_selection())}

    def _get_arg_parser(self, doc_parser=False):
        description = textwrap.dedent(__doc__)

        examples = textwrap.dedent('''
            mdt-create-mask data.nii.gz data.prtcl
            mdt-create-mask data.nii.gz data.prtcl -o data_mask.nii.gz
            mdt-create-mask data.nii.gz data.prtcl -o data_mask.nii.gz --median-radius 2
        ''')
        epilog = self._format_examples(doc_parser, examples)

        parser = argparse.ArgumentParser(description=description, epilog=epilog,
                                         formatter_class=argparse.RawTextHelpFormatter)
        parser.add_argument('dwi',
                            action=mdt.lib.shell_utils.get_argparse_extension_checker(['.nii', '.nii.gz', '.hdr', '.img']),
                            help='the diffusion weighted image').completer = FilesCompleter(['nii', 'gz', 'hdr', 'img'],
                                                                                            directories=False)
        parser.add_argument('protocol',
                            action=mdt.lib.shell_utils.get_argparse_extension_checker(['.prtcl']),
                            help='the protocol file, see mdt-create-protocol').\
            completer = FilesCompleter(['prtcl'], directories=False)
        parser.add_argument('-o', '--output-name',
                            action=mdt.lib.shell_utils.get_argparse_extension_checker(['.nii', '.nii.gz', '.hdr', '.img']),
                            help='the filename of the output file. Default is <dwi_name>_mask.nii.gz').completer = \
            FilesCompleter(['nii', 'gz', 'hdr', 'img'], directories=False)

        parser.add_argument('--median-radius', type=int, default=4,
                            help="Radius (in voxels) of the applied median filter (default 4).")

        parser.add_argument('--mask-threshold', type=float, default=0,
                            help="Everything below this b-value threshold is masked away (value in s/m^2, default 0)")

        parser.add_argument('--numpass', type=int, default=4,
                            help="Number of pass of the median filter (default 4).")

        parser.add_argument('--dilate', type=int, default=1,
                            help="Number of iterations for binary dilation (default 1).")

        parser.add_argument('--cl-device-ind', type=int, nargs='*', choices=self.available_devices.keys(),
                            help="The index of the device we would like to use. This follows the indices "
                                 "in mdt-list-devices and defaults to the first GPU.")

        return parser

    def run(self, args, extra_args):
        dwi_name = os.path.splitext(os.path.realpath(args.dwi))[0]
        dwi_name = dwi_name.replace('.nii', '')

        if args.output_name:
            output_name = os.path.realpath(args.output_name)
        else:
            output_name = dwi_name + '_mask.nii.gz'

        if args.cl_device_ind:
            if isinstance(args.cl_device_ind, int):
                mot.configuration.set_cl_environments([self.available_devices[args.cl_device_ind]])
            else:
                mot.configuration.set_cl_environments([self.available_devices[ind] for ind in args.cl_device_ind])

            mot.configuration.set_load_balancer(EvenDistribution())

        mdt.create_median_otsu_brain_mask(os.path.realpath(args.dwi), os.path.realpath(args.protocol),
                                          output_name, median_radius=args.median_radius,
                                          numpass=args.numpass, dilate=args.dilate,
                                          mask_threshold=args.mask_threshold)

        logger = logging.getLogger(__name__)
        logger.info('Saved the mask to: {}'.format(output_name))


def get_doc_arg_parser():
    return CreateMask().get_documentation_arg_parser()


if __name__ == '__main__':
    CreateMask().start()



from mdt import CompositeModelTemplate

__author__ = 'Francisco.Lagos'


class S0T1_MI_EPI(CompositeModelTemplate):

    description = '''Inversion recovery model. 
    
    This is made to model the MI-EPI sequence, a multi inversion recovery EPI (Renvall et Al. 2016). 
    The Model is based on Stikov et al.'s three parameter model.
    '''
    model_expression = 'S0 * ExpT1DecIR'
    inits = {'ExpT1DecIR.T1': 3.0}
    upper_bounds = {'ExpT1DecIR.T1': 6.0}


class S0T1GRE(CompositeModelTemplate):

    description = '''Gradient echo model.
        
    Models the unweighted signal (aka. b0) with an extra T1.
    
    This is the classical GRE T1-weighted equation, in which the signal is dependent of TR and flip angle.
    '''
    name = 'S0-T1GRE'
    model_expression = 'S0 * ExpT1DecGRE'


class S0T2_exvivo(CompositeModelTemplate):

    description = 'Models the unweighted signal (aka. b0) with an extra T2.'
    model_expression = 'S0 * ExpT2Dec'
    upper_bounds = {'ExpT2Dec.T2': 0.5}


class S0T2(CompositeModelTemplate):

    description = 'Models the unweighted signal (aka. b0) with an extra T2.'
    model_expression = 'S0 * ExpT2Dec'


class MPM(CompositeModelTemplate):

    description = 'Model for estimating biological microstructure of the tissue/sample.'
    model_expression = 'S0 * MPM_Fit'
    likelihood_function = 'Gaussian'



import collections
import copy

__author__ = 'Robbert Harms'
__date__ = "2016-11-10"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class DeferredActionDict(collections.MutableMapping):

    def __init__(self, func, items, cache=True):
        """Applies the given function on the given items at the moment of data request.

        On the moment one of the keys of this dict class is requested we apply the given function on the given items
        and return the result of that function. The advantage of this class is that it defers an expensive operation
        until it is needed.

        Items added to this dictionary after creation are assumed to be final, that is, we won't run the
        function on them.

        Args:
            func (Function): the callback function to apply on the given items at request, with signature:

                .. code-block:: python

                    def callback(key, value)

            items (collections.MutableMapping): the items on which we operate
            cache (boolean): if we want to cache computed results
        """
        self._func = func
        self._items = copy.copy(items)
        self._applied_on_key = {}
        self._cache = cache

    def __delitem__(self, key):
        if key in self._items:
            del self._items[key]
        if key in self._applied_on_key:
            del self._applied_on_key[key]

    def __getitem__(self, key):
        if key not in self._applied_on_key or not self._applied_on_key[key]:
            item = self._func(key, self._items[key])

            if not self._cache:
                return item

            self._items[key] = item
            self._applied_on_key[key] = True
        return self._items[key]

    def __contains__(self, key):
        return key in self._items

    def __iter__(self):
        for key in self._items.keys():
            yield key

    def __len__(self):
        return len(self._items)

    def __setitem__(self, key, value):
        self._items[key] = value
        self._applied_on_key[key] = True

    def __copy__(self):
        new_one = type(self)(self._func, copy.copy(self._items), cache=self._cache)
        new_one._applied_on_key = copy.copy(self._applied_on_key)
        return new_one


class DeferredFunctionDict(collections.MutableMapping):

    def __init__(self, items, cache=True):
        """The items should contain a list of functions that we apply at the moment of request.

        On the moment one of the keys of this dict class is requested we apply the function stored in the items dict
        for that key and return the result of that function.
        The advantage of this class is that it defers an expensive operation until it is needed.

        Items set to this dictionary are assumed to be final, that is, we won't run the function on them.

        Args:
            items (collections.MutableMapping): the items on which we operate, each value should
                contain a function with no parameters that we run to return the results.
            cache (boolean): if we want to cache computed results
        """
        self._items = copy.copy(items)
        self._applied_on_key = {}
        self._cache = cache

    def __delitem__(self, key):
        if key in self._items:
            del self._items[key]
        if key in self._applied_on_key:
            del self._applied_on_key[key]

    def __getitem__(self, key):
        if key not in self._applied_on_key or not self._applied_on_key[key]:
            item = self._items[key]()

            if not self._cache:
                return item

            self._items[key] = item
            self._applied_on_key[key] = True
        return self._items[key]

    def __contains__(self, key):
        return key in self._items

    def __iter__(self):
        for key in self._items.keys():
            yield key

    def __len__(self):
        return len(self._items)

    def __setitem__(self, key, value):
        self._items[key] = value
        self._applied_on_key[key] = True

    def __copy__(self):
        new_one = type(self)(copy.copy(self._items), cache=self._cache)
        new_one._applied_on_key = copy.copy(self._applied_on_key)
        return new_one


class DeferredActionTuple(collections.Sequence):

    def __init__(self, func, items, cache=True):
        """Applies the given function on the given items at moment of request.

        On the moment one of the elements is requested we apply the given function on the given items
        and return the result of that function. The advantage of this class is that it defers an expensive operation
        until it is needed.

        Args:
            func (Function): the callback function to apply on the given items at request, with signature:

                .. code-block:: python

                    def callback(index, value)

            items (list, tuple): the items on which we operate
            cache (boolean): if we want to cache computed results
        """
        self._func = func
        self._items = list(copy.copy(items))
        self._applied_on_index = {}
        self._cache = cache

    def __getitem__(self, index):
        if index not in self._applied_on_index or not self._applied_on_index[index]:
            item = self._func(index, self._items[index])

            if not self._cache:
                return item

            self._items[index] = item
            self._applied_on_index[index] = True
        return self._items[index]

    def __len__(self):
        return len(self._items)

    def __copy__(self):
        new_one = type(self)(self._func, copy.copy(self._items), cache=self._cache)
        new_one._applied_on_index = copy.copy(self._applied_on_index)
        return new_one



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# mdt documentation build configuration file, created by
# sphinx-quickstart on Tue Jul  9 22:26:36 2013.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os
from datetime import datetime
from functools import wraps
from unittest.mock import MagicMock
import builtins
import glob
import re
from textwrap import dedent


mock_as_class = ['QWidget', 'QMainWindow', 'QDialog', 'QObject',
                 'CLRoutine', 'InputData', 'NumDiffInfo', 'SampleModelInterface',
                 'NumericalDerivativeInterface', 'CLFunction', 'SimpleCLFunction']
mock_as_decorator = ['pyqtSlot']
mock_modules = ['mot', 'pyopencl', 'PyQt5', 'matplotlib', 'mpl_toolkits']


def mock_decorator(*args, **kwargs):
    """Mocked decorator, needed in the case we need to mock a decorator"""
    def _called_decorator(dec_func):
        @wraps(dec_func)
        def _decorator(*args, **kwargs):
            return dec_func()
        return _decorator
    return _called_decorator


class MockNamedComponent(MagicMock):
    """Some of the loaded components require the __name__ property to be set, this mock class makes that so."""
    @classmethod
    def __name__(cls):
        return MagicMock()


class MockModule(MagicMock):
    """The base mocking class. This mimics a module."""
    @classmethod
    def __getattr__(cls, name):
        if name in mock_as_class:
            class MockClass(MagicMock):
                @classmethod
                def __getattr__(cls, name):
                    return MockModule()
            return MockClass
        if name in mock_as_decorator:
            return mock_decorator
        return MockNamedComponent()


orig_import = __import__


def import_mock(name, *args, **kwargs):
    """Mock all modules starting with one of the mock_modules names."""
    if any(name.startswith(s) for s in mock_modules):
        return MockModule()
    return orig_import(name, *args, **kwargs)

builtins.__import__ = import_mock


def get_cli_doc_items():
    items = []

    for file in sorted(glob.glob('../mdt/cli_scripts/*.py')):
        module_name = os.path.splitext(os.path.basename(file))[0]
        command_name = module_name.replace('_', '-')

        def get_command_class_name():
            with open(file) as f:
                match = re.search(r'class (\w*)\(', f.read())
                if match:
                    return match.group(1)
                return None

        command_class_name = get_command_class_name()

        if command_class_name is not None:
            item = dedent("""
                .. _cli_index_{command_name}:

                {command_name}
                {command_name_highlight}

                .. argparse::
                   :ref: mdt.cli_scripts.{module_name}.get_doc_arg_parser
                   :prog: {command_name}
            """).format(command_name=command_name, command_name_highlight='='*len(command_name),
                        module_name=module_name)

            items.append(item)
    return items


with open('auto_gen_cli_index.rst', 'w') as f:
    for item in get_cli_doc_items():
        f.write(item[1:] + '\n\n\n')


# If extensions (or modules to document with autodoc) are in another
# directory, add these directories to sys.path here. If the directory is
# relative to the documentation root, use os.path.abspath to make it
# absolute, like shown here.
#sys.path.insert(0, os.path.abspath('..'))

# Building from inside the docs/ directory?
if os.path.basename(os.getcwd()) == 'docs':
    sys.path.insert(1, os.path.abspath(os.path.join('..')))


os.environ["MDT.LOAD_COMPONENTS"] = "0"
import mdt

# -- General configuration ---------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc', 'sphinx.ext.viewcode', 'sphinx.ext.napoleon', 'sphinx.ext.intersphinx',
              'sphinx.ext.mathjax', 'sphinxcontrib.bibtex', 'sphinxarg.ext']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'MDT'
year = datetime.now().year
copyright = u'%d Robbert Harms' % year

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
version = mdt.__version__
# The full version, including alpha/beta/rc tags.
release = mdt.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to
# some non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build', '_getting_started', '_dynamic_modules', 'auto_gen_cli_index.rst']

# The reST default role (used for this markup: `text`) to use for all
# documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ['mdt.']

# If true, keep warnings as "system message" paragraphs in the built
# documents.
#keep_warnings = False

# map to other projects
intersphinx_mapping = {
    'mot': ('http://mot.readthedocs.io/en/latest/', None),
}


# -- Options for HTML output -------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'alabaster'

# Theme options are theme-specific and customize the look and feel of a
# theme further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    'show_powered_by': False,
    'description': "Microstructure Diffusion Toolbox",
    'logo_name': True,
    'sidebar_collapse': True,
    'fixed_sidebar': False,
    'extra_nav_links': {'Module index': 'py-modindex.html'}
}


# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as
# html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the
# top of the sidebar.
#html_logo = None

# The name of an image file (within the static path) to use as favicon
# of the docs.  This file should be a Windows icon file (.ico) being
# 16x16 or 32x32 pixels large.
html_favicon = '_static/html_favicon.ico'

# Add any paths that contain custom static files (such as style sheets)
# here, relative to this directory. They are copied after the builtin
# static files, so a file named "default.css" will overwrite the builtin
# "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page
# bottom, using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {
   '**': ['about.html',
          'navigation.html',
          'searchbox.html']
}

# Additional templates that should be rendered to pages, maps page names
# to template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer.
# Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer.
# Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages
# will contain a <link> tag referring to it.  The value of this option
# must be the base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'mdtdoc'


# -- Options for LaTeX output ------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    'papersize': 'a4paper',

    # The font size ('10pt', '11pt' or '12pt').
    'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    'preamble': r"""
        \makeatletter
        \def\@makechapterhead#1{%
          %%%%\vspace*{50\p@}% %%% removed!
          {\parindent \z@ \raggedright \normalfont
            \ifnum \c@secnumdepth >\m@ne
                \huge\bfseries \@chapapp\space \thechapter
                \par\nobreak
                \vskip 20\p@
            \fi
            \interlinepenalty\@M
            \Huge \bfseries #1\par\nobreak
            \vskip 40\p@
          }}
        \def\@makeschapterhead#1{%
          %%%%%\vspace*{50\p@}% %%% removed!
          {\parindent \z@ \raggedright
            \normalfont
            \interlinepenalty\@M
            \Huge \bfseries  #1\par\nobreak
            \vskip 40\p@
          }}
        \makeatother

        \setcounter{secnumdepth}{1}
        
        \usepackage{titlesec}
        \titlespacing*{\section}{0pt}{6ex plus 1ex minus .2ex}{1ex plus .1ex}
        \titlespacing*{\subsection}{0pt}{4ex plus 1ex minus .2ex}{0ex plus .1ex}
        \titlespacing*{\subsubsection}{0pt}{3ex plus 1ex minus .2ex}{0ex}
    """,
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_documents = [
    ('index_latex', 'mdt.tex',
     u'Microstructure Diffusion Toolbox',
     u'Robbert Harms', 'manual'),
]

# The name of an image file (relative to this directory) to place at
# the top of the title page.
latex_logo = '../mdt/data/logo_docs.png'

# For "manual" documents, if this is true, then toplevel headings
# are parts, not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index_man', 'mdt',
     u'Microstructure Diffusion Toolkit',
     [u'Robbert Harms'], 1)
]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output ----------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    ('index_latexz', 'mdt',
     u'Microstructure Diffusion Toolbox',
     u'Robbert Harms',
     'mdt',
     'One line description of project.',
     'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False

# -- Options for napoleon ----
autoclass_content = 'both'



from mdt import CompartmentTemplate, FreeParameterTemplate

__author__ = 'Francisco Fritz'
__licence__ = 'LGPL v3'


class ExpT1DecGRE(CompartmentTemplate):

    parameters = ('TR', 'flip_angle', 'excitation_b1', 'T1')
    cl_code = """
        return sin((double)flip_angle * excitation_b1) *
              (1 - exp(-TR / (double)T1)) /
              (1 - cos((double)flip_angle * excitation_b1) * exp(-TR / (double)T1) );
    """


class ExpT1DecIR(CompartmentTemplate):
    description = """IR equation.

    IR equation in which TI and TR are considered to estimate T1. Assuming TE << T1, the TE component of the signal
    is discarded. In cascade, S0 contains T2 and PD weighted information. An efficiency factor is added to the TI
    parameter.
    
    This is made to model the MI-EPI sequence, a multi inversion recovery epi (Renvall et Al. 2016). 
    The Model is based on Stikov et al.'s three parameter model.
    """
    parameters = ('TR', 'TI', 'Efficiency', 'T1')
    cl_code = """
        return fabs(1 + exp(-TR / (double)T1) - 2 * Efficiency * exp(-TI / (double)T1));
    """

    class Efficiency(FreeParameterTemplate):
        init_value = 0.95
        lower_bound = 0
        upper_bound = 1
        parameter_transform = 'SinSqrClamp'
        sampling_proposal_std = 0.001


class ExpT1DecTM(CompartmentTemplate):

    parameters = ('SEf', 'TR', 'TM', 'TE', 'T1', 'flip_angle', 'excitation_b1', 'Refoc_fa1',
                  'refocusing1_b1', 'Refoc_fa2', 'refocusing2_b1', 'b', 'd_exvivo')
    cl_code = """
        return powr((double)0.5, (double)SEf)
            * sin((double)flip_angle * excitation_b1)
            * sin((double)Refoc_fa1 * refocusing1_b1)
            * sin((double)Refoc_fa2 * (refocusing2_b1 * SEf + refocusing1_b1 * (1 - SEf)))
            * (1 - exp(-(TR - TM) / (double)T1))
            * exp(- ((TM * SEf) / (double)T1) - (double)(b * d_exvivo));
    """


class ExpT1DecTM_simple(CompartmentTemplate):

    parameters = ('TM', 'T1')
    cl_code = 'return exp(-TM / T1);'


class ExpT1DecTR(CompartmentTemplate):

    parameters = ('TR', 'T1')
    cl_code = 'return abs(1 - exp(-TR / T1));'


class ExpT1ExpT2GRE(CompartmentTemplate):

    parameters = ('TR', 'TE', 'excitation_b1', 'flip_angle', 'T1', 'T2')
    cl_code = """
        return sin(flip_angle * excitation_b1) * (1 - exp(-TR / T1)) 
                / (1 - cos(flip_angle * excitation_b1) * exp(-TR / T1)) * exp(-TE / T2);
    """


class ExpT1ExpT2sGRE(CompartmentTemplate):

    parameters = ('TR', 'TE', 'excitation_b1', 'flip_angle', 'T1', 'T2s')
    cl_code = """
        return sin(flip_angle * excitation_b1) * (1 - exp(-TR / T1)) 
                / (1 - cos(flip_angle * excitation_b1) * exp(-TR / T1)) * exp(-TE / T2s);
    """


class ExpT1ExpT2STEAM(CompartmentTemplate):
    description = """Generalised STEAM equation.

    From protocol, if the signal is SE, we can setup TM = 0 in all the volumes,
    which returns to the standard SE signal decay

    This equation can be used to calculate relaxation time (T1/T2) from spin echo (SE) and/or stimulated spin echo (STE)
    data. It is important to notice that in the protocol you have to define some parameters in a specific way:

    (1) For SE data, the original equation contains only the first refocusing pulse variable, but half of this value
        and in the power of two (sin(Refoc_fa1/2)**2). For that it is needed to define Refoc_fa2 = Refoc_fa1 and
        Refoc_fa1 has to be HALF of the used FA in the protocol (then, also Refoc_fa2). Also, the 0.5 factor is
        not included, then SEf (Spin echo flag) should be 0. Finally, TM (mixing time) has to be 0.
    (2) For STE data, this equation is used totally. Just SEf = 1.

    //UPDATE (24.03.17): This sequence is valid for STE SIGNAL ONLY! Don't mix with SE volumes.
    """
    parameters = ('SEf', 'TR', 'TE', 'TM', 'b', 'flip_angle', 'excitation_b1', 'Refoc_fa1',
                  'refocusing1_b1', 'Refoc_fa2', 'refocusing2_b1', 'T2', 'T1', 'd_exvivo')
    cl_code = """
        return sin((double)flip_angle * (double)excitation_b1)
            *   sin((double)Refoc_fa1 * (double)refocusing1_b1)
            *   sin((double)Refoc_fa2 * (double)refocusing2_b1)
            *   (exp(- (double)TM / (double) T1) - exp(- (double)TR / (double)T1))
            *   exp(- (double)TE / (double)T2)
            *   exp(- (double)b * (double)d_exvivo);
    """


class ExpT2Dec(CompartmentTemplate):

    parameters = ('TE', 'T2')
    cl_code = 'return exp(-TE / T2);'


class ExpT2DecSTEAM(CompartmentTemplate):

    parameters = ('SEf', 'TE', 'TM', 'b', 'flip_angle', 'excitation_b1', 'Refoc_fa1',
                      'refocusing1_b1', 'Refoc_fa2', 'refocusing2_b1', 'T2', 'T1', 'd_exvivo')
    cl_code = """
        return powr((double)0.5, (double)SEf)
            *   sin((double)flip_angle * (double)excitation_b1)
            *   sin((double)Refoc_fa1 * (double)refocusing1_b1)
            *   sin((double)Refoc_fa2 * ((double)refocusing2_b1 * (double)SEf + 
                                         (double)refocusing1_b1 * (double)(1 - SEf)))
            *   exp(- (double)TE / (double)T2)
            *   exp(- (double)TM * SEf / (double) T1)
            *   exp(- (double)b * (double)d_exvivo);
    """


class MPM_Fit(CompartmentTemplate):
    description = """MPM fitting (Weiskopf, 2016 ESMRMB Workshop)

    This fitting is a model published by Helms (2008) and Weiskopf (2011) to determinate biological properties
    of the tissue/sample in function *of several images*, which includes T1w, PDw and MTw images. 
    This function is still an approximation and only if the assumptions of the approximations hold for ex-vivo tissue, 
    then can be used for ex-vivo data.
    """
    parameters = ('TR', 'flip_angle', 'excitation_b1', 'T1')
    cl_code = 'return (flip_angle * excitation_b1) * ( (TR / T1) ' \
              '     / ( pown(flip_angle * excitation_b1, 2) / 2 + ( TR / T1 ) ) );'


class LinMPM_Fit(CompartmentTemplate):
    description = """MPM fitting (Weiskopf, 2016 ESMRMB Workshop)

    This fitting is a model published by Helms (2008) and Weiskopf (2011) to determinate biological properties
    of the tissue/sample in function *of several images*, which includes T1w, PDw and MTw images. 
    This function is still an approximation and only if the assumptions of the approximations hold for ex-vivo tissue, 
    then can be used for ex-vivo data.
    """
    parameters = ('TR', 'flip_angle', 'b1', 'T1')
    cl_code = 'return log(flip_angle * b1) + log(TR / T1)' \
              '       - log( pown(flip_angle * b1, 2) / 2 + ( TR / T1 ) ) ;'


class LinT1GRE(CompartmentTemplate):
    description = """Lineal T1 fitting (Weiskopf, 2016 ESMRMB Workshop)

    This fitting is the extension of the standard GRE equation for flip angles lower than 90deg. This modelling allows a
    linear fitting of the data if is enough data to support it. In principle, it should not be a problem if only two
    points are used, however the addition of a constant in the equation could give some kind of uncertainty.

    B1 has to be normalized *in function of the reference voltage, the angle distribution and the reference angle*.
    Here I assume that TR <<< T1, then exp(-TR/T1) ~ 1 - TR/T1. Then the equation becomes 'simpler'. 
    However, if this condition is not achieved, then return to the standard equation.
    
    Also, DATA HAS TO BE PROCESSED BEFORE TO USE THIS EQUATION. Please apply log() on the data.
    """
    parameters = ('Sw', 'E1')
    cl_code = """
        return Sw * E1;
    """


class LinT2Dec(CompartmentTemplate):

    parameters = ('TE', 'R2')
    cl_code = 'return -TE * R2;'




import glob
import os
from itertools import filterfalse

from mdt.lib.batch_utils import SimpleBatchProfile, BatchFitProtocolLoader, SimpleSubjectInfo
from mdt.component_templates.base import ComponentBuilder, ComponentTemplate

__author__ = 'Robbert Harms'
__date__ = "2017-02-14"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class BatchProfileBuilder(ComponentBuilder):

    def _create_class(self, template):
        """Creates classes with as base class SimpleBatchProfile

        Args:
            template (BatchProfileTemplate): the batch profile config template

        Returns:
            SimpleBatchProfile: an subclass of a batch profile
        """
        class AutoCreatedBatchProfile(SimpleBatchProfile):
            def _get_subjects(self, data_folder):
                dirs = sorted([os.path.basename(f) for f in glob.glob(os.path.join(data_folder, '*'))])

                subjects = []
                for subject_id in dirs:
                    subject_base_folder = os.path.join(
                        data_folder, template.subject_base_folder.format(subject_id=subject_id))

                    def _prepare_path(template_path):
                        if template_path is None:
                            return None
                        return template_path.format(data_folder=data_folder,
                                                    subject_id=subject_id,
                                                    subject_base_folder=subject_base_folder)

                    data_glob = glob.glob(_prepare_path(template.data_fname))
                    if not list(data_glob):
                        break

                    noise_std = self._autoload_noise_std(data_folder, subject_id, file_pattern=template.noise_std_fname)

                    protocol_loader = BatchFitProtocolLoader(
                        _prepare_path(template.protocol_auto_dir),
                        protocol_fname=_prepare_path(template.protocol_fname),
                        bvec_fname=_prepare_path(template.bvec_fname),
                        bval_fname=_prepare_path(template.bval_fname),
                        protocol_columns=template.protocol_columns)

                    mask_fname = None
                    if list(glob.glob(_prepare_path(template.mask_fname))):
                        mask_fname = glob.glob(_prepare_path(template.mask_fname))[0]
                        data_glob = list(filterfalse(lambda v: v == mask_fname, data_glob))

                    grad_dev = None
                    if list(glob.glob(_prepare_path(template.gradient_deviations_fname))):
                        grad_dev = glob.glob(_prepare_path(template.gradient_deviations_fname))[0]
                        data_glob = list(filterfalse(lambda v: v == grad_dev, data_glob))

                    subjects.append(SimpleSubjectInfo(
                        subject_base_folder, subject_id, data_glob[0],
                        protocol_loader, mask_fname, noise_std=noise_std,
                        gradient_deviations=grad_dev))

                return subjects

            def __str__(self):
                return template.name

        for name, method in template.bound_methods.items():
            setattr(AutoCreatedBatchProfile, name, method)

        return AutoCreatedBatchProfile


class BatchProfileTemplate(ComponentTemplate):
    """The batch profile template to inherit.

    Attributes:
        name (str): the name of this batch profile
        description (str): the description
        subject_base_folder (str): the base folder for this subject. Allows expansion of ``{subject_id}``.
        data_fname (str): the filename of the data volumes file. Allows expansion of ``{subject_id}``
            and ``{subject_base_folder}``, and supports globbing. Results are afterwards filtered to exclude matches
            of the mask and gradient deviations files.
        mask_fname (str): the filename of the mask. Allows expansion of ``{subject_id}`` and ``{subject_base_folder}``,
            and supports globbing.
        noise_std_fname (str): the filename of the noise standard deviation file. Can be textfile or a nifti file.
            Allows expansion of ``{subject_id}`` and ``{subject_base_folder}``, and supports globbing.
        gradient_deviations_fname (str): the filename of the gradient deviations. Allows expansion of
            ``{subject_id}`` and ``{subject_base_folder}``, and supports globbing.
        protocol_auto_dir (str): the directory from which MDT will try to autoload a directory.
            Supports ``{subject_id}`` and ``{subject_base_folder}``, and supports globbing.
        protocol_fname (str): the filename of the protocol file to use. Supports ``{subject_id}``
            and ``{subject_base_folder}``, and supports globbing. If provided, we use it instead of the automatically
            searched default.
        bvec_fname (str): the filename of the bvec file to use. Supports ``{subject_id}``
            and ``{subject_base_folder}``, and supports globbing. If provided, we use it instead of the automatically
            searched default.
        bval_fname (str): the filename of the bval file to use. Supports ``{subject_id}``
            and ``{subject_base_folder}``, and supports globbing. If provided, we use it instead of the automatically
            searched default.
        protocol_columns (dict): a dictionary with additional columns to add to the protocol file. Use this for
            default values for all subjects in your study.
    """
    _component_type = 'batch_profiles'
    _builder = BatchProfileBuilder()

    name = None
    description = ''

    subject_base_folder = '{subject_id}'

    data_fname = '{subject_base_folder}/*.nii*'
    mask_fname = '{subject_base_folder}/*mask.nii*'
    noise_std_fname = '{subject_base_folder}/noise_std*'
    gradient_deviations_fname = '{subject_base_folder}/grad_dev.nii*'

    protocol_auto_dir = '{subject_base_folder}'
    protocol_fname = None
    bvec_fname = None
    bval_fname = None
    protocol_columns = {}



from mdt import LibraryFunctionTemplate

__author__ = 'Robbert Harms'
__date__ = '2018-07-11'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class BesselRoots(LibraryFunctionTemplate):
    description = '''
        Constant arrays holding some often used Bessel roots. 
    '''
    is_function = False
    cl_code = '''
        /** 
         * Zeros of integer-order Bessel function derivative Jn'(x).
           Generated using the Python code:
           
           from scipy.special import jnp_zeros
           print(jnp_zeros(1, 16))
         
         */
        __constant int bessel_roots_jnp_length = 16;
        __constant mot_float_type bessel_roots_jnp[] = {
            1.8411837813406593, 5.3314427735250325, 8.536316366346286,  11.706004902592063, 
            14.863588633909032, 18.015527862681804, 21.16436985918879,  24.311326857210776, 
            27.457050571059245, 30.601922972669094, 33.746182898667385, 36.88998740923681, 
            40.03344405335068,  43.17662896544882,  46.319597561173914, 49.46239113970275
        };
    '''



from PyQt5.QtWidgets import QDoubleSpinBox

__author__ = 'Robbert Harms'
__date__ = "2017-01-26"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class QDoubleSpinBoxDotSeparator(QDoubleSpinBox):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def valueFromText(self, text):
        try:
            return float(text)
        except ValueError:
            return 0

    def textFromValue(self, value):
        return str(value)



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""Evaluate an expression on a set of images.

This is meant to quickly convert/combine one or two maps with a mathematical expression.
The expression can be any valid python expression.

The input list of images are loaded as numpy arrays and stored in the array 'input' and 'i'.
Next, the expression is evaluated using the input images and the result is stored in the indicated file(s).

In the expression you can either use the arrays 'input' or 'i' with linear indices, or/and you can use alphabetic
characters for each image. For example, if you have specified 2 input images
you can address them as:

    - input[0] or i[0] or a
    - input[1] or i[1] or b

This linear alphabetic indexing works with every alphabetic character except for the 'i' since that
one is reserved for the array.

The module numpy is available under 'np' and some functions of MDT under 'mdt'.
This allows expressions like::

    np.mean(np.concatenate(i, axis=3), axis=3)

to get the mean value per voxel of all the input images.

It is possible to change the mode of evaluation from single expression to a more complex python
statement using the switch --as-statement (the default is --as-expression). In a statement
more complex python commands are allowed. In statement mode you must explicitly output the
results using 'return'. (Basically it wraps your command in a function, of which the output is
used as expression value).

If no output file is specified and the output is of dimension 2 or lower we print the output directly
to the console. If the output is a list or tuple instead of an ndarray we will write every element of that
sequence as a separate file. This will use the indicated file as basename and append an numerical index to it.
"""
import argparse
import glob
import os
from collections import Sequence
import numpy as np

import mdt
from argcomplete.completers import FilesCompleter
import textwrap

from mdt.lib.shell_utils import BasicShellApplication
from mdt.utils import split_image_path

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class MathImg(BasicShellApplication):

    def _get_arg_parser(self, doc_parser=False):
        description = textwrap.dedent(__doc__)

        examples = textwrap.dedent('''
            mdt-math-img fiso.nii ficvf.nii '(1-input[0]) * i[1]' -o Wic.w.nii.gz
            mdt-math-img fiso.nii ficvf.nii '(1-a) * b' -o Wic.w.nii.gz
            mdt-math-img *.nii.gz 'np.mean(np.concatenate(i, axis=3), axis=3)' -o output.nii.gz
            mdt-math-img FA.nii.gz 'np.mean(a)'
            mdt-math-img FA.nii white_matter_mask.nii 'np.mean(mdt.create_roi(a, b))'
            mdt-math-img images*.nii.gz mask.nii 'list(map(lambda f: np.mean(mdt.create_roi(f, i[-1])), i[0:-1]))'
            mdt-math-img FA.nii.gz
            mdt-math-img timeseries.nii '[a[..., ind] for ind in range(a.shape[-1])]' -o split.nii
           ''')
        epilog = self._format_examples(doc_parser, examples)

        parser = argparse.ArgumentParser(description=description, epilog=epilog,
                                         formatter_class=argparse.RawTextHelpFormatter)

        parser.add_argument('input_files', metavar='input_files', nargs="+", type=str,
                            help="The input images to use")

        parser.add_argument('expr', metavar='expr', type=str,
                            help="The expression/statement to evaluate.")

        parser.add_argument('-e', '--as-expression', dest='as_expression', action='store_true',
                            help="Evaluates the given string as an expression (default).")
        parser.add_argument('-s', '--as-statement', dest='as_expression', action='store_false',
                            help="Evaluates the given string as an statement.")
        parser.set_defaults(as_expression=True)

        parser.add_argument('-o', '--output-file',
                            help='the output file, if not set nothing is written').completer = \
            FilesCompleter(['nii', 'gz', 'hdr', 'img'], directories=False)

        parser.add_argument('-4d', '--input-4d', action='store_true',
                            help='Add a singleton dimension to all input 3d maps to make them 4d, this prevents '
                                 'some broadcast issues.')

        parser.add_argument('--verbose', '-v', action='store_true', help="Verbose, prints runtime information")

        return parser

    def run(self, args, extra_args):
        write_output = args.output_file is not None

        file_names = []
        images = []
        for file in args.input_files:
            globbed = glob.glob(file)

            if globbed:
                for fname in globbed:
                    file_names.append(fname)
                    images.append(mdt.load_nifti(os.path.realpath(fname)).get_data())
            else:
                file_names.append(file)
                images.append(mdt.load_nifti(os.path.realpath(file)).get_data())

        if args.verbose:
            print('')

        if args.input_4d:
            images = self._images_3d_to_4d(images)

        context_dict = {'input': images, 'i': images, 'np': np, 'mdt': mdt}
        alpha_chars = list('abcdefghjklmnopqrstuvwxyz')

        for ind, image in enumerate(images):
            context_dict.update({alpha_chars[ind]: image})

            if args.verbose:
                print('Input {ind} ({alpha}):'.format(ind=ind, alpha=alpha_chars[ind]))
                print('    name: {}'.format(split_image_path(file_names[ind])[1]))
                print('    shape: {}'.format(str(image.shape)))

        if args.verbose:
            print('')
            print("Evaluating: '{expr}'".format(expr=args.expr))

        if args.as_expression:
            output = eval(args.expr, context_dict)
        else:
            expr = textwrap.dedent('''
            def mdt_image_math():
                {}
            output = mdt_image_math()
            ''').format(args.expr)
            exec(expr, context_dict)
            output = context_dict['output']

        if args.verbose:
            print('')
            if isinstance(output, np.ndarray):
                print('Output shape: {shape}'.format(shape=str(output.shape)))
            else:
                print('Output is single value')

            print('Output: ')
            print('')
            print(output)
        else:
            if not write_output:
                print(output)

        if args.verbose:
            print('')

        if write_output:
            output_file = os.path.realpath(args.output_file)
            dirname, basename, ext = split_image_path(output_file)

            if isinstance(output, Sequence):
                for ind, element in enumerate(output):
                    mdt.write_nifti(element,
                                    dirname + basename + '_' + str(ind) + ext,
                                    mdt.load_nifti(file_names[0]).get_header())
            else:
                mdt.write_nifti(output, output_file, mdt.load_nifti(file_names[0]).get_header())

    def _images_3d_to_4d(self, images):
        return list([image[..., np.newaxis] if len(image.shape) == 3 else image for image in images])


def get_doc_arg_parser():
    return MathImg().get_documentation_arg_parser()


if __name__ == '__main__':
    MathImg().start()



from mdt import CascadeTemplate

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class BallStick_r1(CascadeTemplate):

    description = 'Cascade for Ballstick'
    models = ('S0',
              'BallStick_r1')


class BallStick_r2(CascadeTemplate):

    description = 'Cascade for BallStick_r2.'
    models = ('BallStick_r1 (Cascade)',
              'BallStick_r2')
    inits = {'BallStick_r2': {'w_stick1.w': 0.0}}


class BallStick_r3(CascadeTemplate):

    description = 'Cascade for BallStick_r3.'
    models = ('BallStick_r2 (Cascade)',
              'BallStick_r3')
    inits = {'BallStick_r3': {'w_stick2.w': 0.0}}



from mdt import LibraryFunctionTemplate

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class MRIConstants(LibraryFunctionTemplate):

    description = 'Some constants that might be of use in dMRI model functions'
    is_function = False
    cl_code = '''
        /**
         * Gamma represent the gyromagnetic ratio of protons in water (nucleus of H)
         * and are in units of (rad s^-1 T^-1)
         **/
        #ifndef GAMMA_H
            #define GAMMA_H 267.5987E6
        #endif

        /**
         * The same GAMMA as defined above but then divided by 2*pi, in units of s^-1 T^-1
         **/
        #ifndef GAMMA_H_HZ
            #define GAMMA_H_HZ (GAMMA_H / (2 * M_PI))
        #endif

        /**
         * The square of the original GAMMA defined above
         */
        #ifndef GAMMA_H_SQ
            #define GAMMA_H_SQ (GAMMA_H * GAMMA_H)
        #endif

        /**
         * The square of the GAMMA divided by 2*pi from above
         */
        #ifndef GAMMA_H_HZ_SQ
            #define GAMMA_H_HZ_SQ (GAMMA_H_HZ * GAMMA_H_HZ)
        #endif
    '''



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'optimization_extra_data_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.5.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_OptimizationExtraDataDialog(object):
    def setupUi(self, OptimizationExtraDataDialog):
        OptimizationExtraDataDialog.setObjectName("OptimizationExtraDataDialog")
        OptimizationExtraDataDialog.resize(843, 331)
        self.verticalLayout = QtWidgets.QVBoxLayout(OptimizationExtraDataDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.label_3 = QtWidgets.QLabel(OptimizationExtraDataDialog)
        font = QtGui.QFont()
        font.setPointSize(14)
        self.label_3.setFont(font)
        self.label_3.setObjectName("label_3")
        self.verticalLayout_3.addWidget(self.label_3)
        self.label_4 = QtWidgets.QLabel(OptimizationExtraDataDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_4.setFont(font)
        self.label_4.setObjectName("label_4")
        self.verticalLayout_3.addWidget(self.label_4)
        self.verticalLayout.addLayout(self.verticalLayout_3)
        self.line = QtWidgets.QFrame(OptimizationExtraDataDialog)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.label_15 = QtWidgets.QLabel(OptimizationExtraDataDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_15.setFont(font)
        self.label_15.setObjectName("label_15")
        self.gridLayout.addWidget(self.label_15, 2, 2, 1, 1)
        self.label_10 = QtWidgets.QLabel(OptimizationExtraDataDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_10.setFont(font)
        self.label_10.setObjectName("label_10")
        self.gridLayout.addWidget(self.label_10, 4, 2, 1, 1)
        self.line_5 = QtWidgets.QFrame(OptimizationExtraDataDialog)
        self.line_5.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_5.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_5.setObjectName("line_5")
        self.gridLayout.addWidget(self.line_5, 3, 0, 1, 3)
        self.label_7 = QtWidgets.QLabel(OptimizationExtraDataDialog)
        self.label_7.setObjectName("label_7")
        self.gridLayout.addWidget(self.label_7, 0, 0, 1, 1)
        self.label_14 = QtWidgets.QLabel(OptimizationExtraDataDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_14.setFont(font)
        self.label_14.setObjectName("label_14")
        self.gridLayout.addWidget(self.label_14, 0, 2, 1, 1)
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.gradientDeviations = QtWidgets.QLineEdit(OptimizationExtraDataDialog)
        self.gradientDeviations.setObjectName("gradientDeviations")
        self.horizontalLayout_6.addWidget(self.gradientDeviations)
        self.gradDevFileSelect = QtWidgets.QPushButton(OptimizationExtraDataDialog)
        self.gradDevFileSelect.setObjectName("gradDevFileSelect")
        self.horizontalLayout_6.addWidget(self.gradDevFileSelect)
        self.gridLayout.addLayout(self.horizontalLayout_6, 2, 1, 1, 1)
        self.label_8 = QtWidgets.QLabel(OptimizationExtraDataDialog)
        self.label_8.setObjectName("label_8")
        self.gridLayout.addWidget(self.label_8, 2, 0, 1, 1)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.noiseStd = QtWidgets.QLineEdit(OptimizationExtraDataDialog)
        self.noiseStd.setObjectName("noiseStd")
        self.horizontalLayout_5.addWidget(self.noiseStd)
        self.noiseStdFileSelect = QtWidgets.QPushButton(OptimizationExtraDataDialog)
        self.noiseStdFileSelect.setObjectName("noiseStdFileSelect")
        self.horizontalLayout_5.addWidget(self.noiseStdFileSelect)
        self.gridLayout.addLayout(self.horizontalLayout_5, 0, 1, 1, 1)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setContentsMargins(-1, 0, -1, -1)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.protocolMaps = QtWidgets.QListWidget(OptimizationExtraDataDialog)
        self.protocolMaps.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)
        self.protocolMaps.setObjectName("protocolMaps")
        self.verticalLayout_2.addWidget(self.protocolMaps)
        self.horizontalLayout_8 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_8.setObjectName("horizontalLayout_8")
        self.addProtocolMap = QtWidgets.QPushButton(OptimizationExtraDataDialog)
        self.addProtocolMap.setObjectName("addProtocolMap")
        self.horizontalLayout_8.addWidget(self.addProtocolMap)
        self.removeProtocolMap = QtWidgets.QPushButton(OptimizationExtraDataDialog)
        self.removeProtocolMap.setObjectName("removeProtocolMap")
        self.horizontalLayout_8.addWidget(self.removeProtocolMap)
        self.verticalLayout_2.addLayout(self.horizontalLayout_8)
        self.gridLayout.addLayout(self.verticalLayout_2, 4, 1, 1, 1)
        self.line_2 = QtWidgets.QFrame(OptimizationExtraDataDialog)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.gridLayout.addWidget(self.line_2, 1, 0, 1, 3)
        self.label_2 = QtWidgets.QLabel(OptimizationExtraDataDialog)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 4, 0, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.line_3 = QtWidgets.QFrame(OptimizationExtraDataDialog)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.verticalLayout.addWidget(self.line_3)
        self.buttonBox = QtWidgets.QDialogButtonBox(OptimizationExtraDataDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(OptimizationExtraDataDialog)
        self.buttonBox.accepted.connect(OptimizationExtraDataDialog.accept)
        self.buttonBox.rejected.connect(OptimizationExtraDataDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(OptimizationExtraDataDialog)
        OptimizationExtraDataDialog.setTabOrder(self.noiseStd, self.noiseStdFileSelect)
        OptimizationExtraDataDialog.setTabOrder(self.noiseStdFileSelect, self.gradientDeviations)
        OptimizationExtraDataDialog.setTabOrder(self.gradientDeviations, self.gradDevFileSelect)
        OptimizationExtraDataDialog.setTabOrder(self.gradDevFileSelect, self.protocolMaps)
        OptimizationExtraDataDialog.setTabOrder(self.protocolMaps, self.addProtocolMap)
        OptimizationExtraDataDialog.setTabOrder(self.addProtocolMap, self.removeProtocolMap)

    def retranslateUi(self, OptimizationExtraDataDialog):
        _translate = QtCore.QCoreApplication.translate
        OptimizationExtraDataDialog.setWindowTitle(_translate("OptimizationExtraDataDialog", "Additional input data"))
        self.label_3.setText(_translate("OptimizationExtraDataDialog", "Additional input data"))
        self.label_4.setText(_translate("OptimizationExtraDataDialog", "Extra data that can be used in the model fitting procedure"))
        self.label_15.setText(_translate("OptimizationExtraDataDialog", "(Per voxel 9 values that constitute the gradient non-linearities)"))
        self.label_10.setText(_translate("OptimizationExtraDataDialog", "(Provide voxel-based maps for the protocol parameters. \n"
"If given, these override the values from the protocol file.)"))
        self.label_7.setText(_translate("OptimizationExtraDataDialog", "Noise standard deviation:"))
        self.label_14.setText(_translate("OptimizationExtraDataDialog", "(Empty for auto detection, or set a scalar or a path to a nifti file)"))
        self.gradDevFileSelect.setText(_translate("OptimizationExtraDataDialog", "File browser"))
        self.label_8.setText(_translate("OptimizationExtraDataDialog", "Gradient deviations:"))
        self.noiseStdFileSelect.setText(_translate("OptimizationExtraDataDialog", "File browser"))
        self.addProtocolMap.setText(_translate("OptimizationExtraDataDialog", "Add map"))
        self.removeProtocolMap.setText(_translate("OptimizationExtraDataDialog", "Remove selected"))
        self.label_2.setText(_translate("OptimizationExtraDataDialog", "Protocol maps:"))




from mdt import CompartmentTemplate

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CylinderGPD(CompartmentTemplate):

    parameters = ('g', 'b', 'G', 'Delta', 'delta', 'd', 'theta', 'phi', 'R')
    dependencies = ('VanGelderenCylinder', 'SphericalToCartesian')
    cl_code = '''
        const mot_float_type direction_2 = pown(dot(g, SphericalToCartesian(theta, phi)), 2);

        const mot_float_type signal_par = -b * d * direction_2;
        const mot_float_type signal_perp = (1 - direction_2) * VanGelderenCylinder(G, Delta, delta, d, R);
        
        return exp(signal_perp + signal_par);
    '''



from mdt import CompartmentTemplate


__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class NODDI_IC(CompartmentTemplate):
    """Generate the compartment model signal for the NODDI Intra Cellular (Stick with dispersion) compartment.

    This is a transcription from the NODDI matlab toolbox, but with a few changes. Most notably, support for the
    cylindrical model has been removed to simplify the code.
    """
    parameters = ('g', 'b', 'd', 'theta', 'phi', 'kappa')
    dependencies = ('erfi',
                    'MRIConstants',
                    'SphericalToCartesian')
    cl_code = '''
        mot_float_type LePerp = 0; // used to be "VanGelderenCylinder(G, Delta, delta, d, R)" with R fixed to 0.
        mot_float_type ePerp = exp(LePerp);
        mot_float_type Lpmp = LePerp + d * b;

        mot_float_type watson_coeff[NODDI_IC_MAX_POLYNOMIAL_ORDER + 1];
        NODDI_IC_WatsonSHCoeff(kappa, watson_coeff);

        mot_float_type lgi[NODDI_IC_MAX_POLYNOMIAL_ORDER + 1];
        NODDI_IC_LegendreGaussianIntegral(Lpmp, lgi);

        // split the summation into two parts to save one array (reusing the lgi array for the legendre terms)
        for(int i = 0; i < NODDI_IC_MAX_POLYNOMIAL_ORDER + 1; i++){
            watson_coeff[i] *= lgi[i] * sqrt((i + 0.25)/M_PI_F);
        }

        NODDI_IC_create_legendre_terms(dot(g, SphericalToCartesian(theta, phi)), lgi);

        mot_float_type signal = 0.0;
        for(int i = 0; i < NODDI_IC_MAX_POLYNOMIAL_ORDER + 1; i++){
            signal += lgi[i] * watson_coeff[i];
        }

        return ePerp * signal / 2.0;
    '''
    cl_extra = '''
        // do not change this value! It would require adding approximations to the functions below
        #define NODDI_IC_MAX_POLYNOMIAL_ORDER 6
        // sqrt(pi)/2
        #define M_SQRTPI_2_F 0.8862269254527580f
        // sqrt(pi)
        #define M_SQRTPI_F 1.7724538509055160f


        /**
         * This will create the legendre terms we need for the NODDI IC model.
         *
         * For the NODDI IC model we need to have a few legendre terms for the same position (argument to x)
         * with linearly increasing degrees of step size 2.
         *
         * This is a specialized version of the function of the function firstLegendreTerm in the MOT library.
         *
         * This function will fill the given array legendre_terms with the values:
         * [0] = firstLegendreTerm(x, 0)
         * [1] = firstLegendreTerm(x, 2)
         * [2] = firstLegendreTerm(x, 4)
         * [3] = firstLegendreTerm(x, 8)
         * ...
         *
         * Changes:
         *  - 2018-06-04, Dr. Luke Edwards: corrected the indices to correctly account for a step size of two.
         */
        void NODDI_IC_create_legendre_terms(const mot_float_type x, mot_float_type* const legendre_terms){
            if(fabs(x) == 1.0){
                // this is the default if fabs(x) == 1.0
                // to eliminate the branch I added this to the front, this saves an if/else.
                // also, since we are after the legendre terms with a list with n = [0, 2, 4, 6, 8, ...]
                // the legendre terms collaps to this loop if fabs(x) == 1.0
                
                for(int i = 0; i < NODDI_IC_MAX_POLYNOMIAL_ORDER + 1; i++){
                    legendre_terms[i] = 1.0;
                }
                return;
            }

            legendre_terms[0] = 1.0;

            mot_float_type P0 = 1.0;
            mot_float_type P1 = x;
            mot_float_type Pn;

            for(int k = 1; k < NODDI_IC_MAX_POLYNOMIAL_ORDER + 1; k++){
                Pn = ((2 * (2*k-1) + 1) * x * P1 - ((2*k-1) * P0)) / ((2*k-1) + 1);
                P0 = P1;
                P1 = Pn;

                legendre_terms[k] = Pn;

                Pn = ((2 * ((2*k-1)+1) + 1) * x * P1 - (((2*k-1)+1) * P0)) / (((2*k-1)+1) + 1);
                P0 = P1;
                P1 = Pn;
            }
        }

        /**
            Copied from the Matlab NODDI toolbox

            function [L, D] = legendreGaussianIntegral(x, n)
            Computes legendre gaussian integrals up to the order specified and the
            derivatives if requested

            The integral takes the following form, in Mathematica syntax,

            L[x, n] = Integrate[Exp[-x \mu^2] Legendre[2*n, \mu], {\mu, -1, 1}]
            D[x, n] = Integrate[Exp[-x \mu^2] (-\mu^2) Legendre[2*n, \mu], {\mu, -1, 1}]

            original author: Gary Hui Zhang (gary.zhang@ucl.ac.uk)
        */
        void NODDI_IC_LegendreGaussianIntegral(const mot_float_type x, mot_float_type* const result){

            if(x > 0.05){
                // exact
                mot_float_type tmp[NODDI_IC_MAX_POLYNOMIAL_ORDER + 1];
                tmp[0] = M_SQRTPI_F * erf(sqrt(x))/sqrt(x);
                for(int i = 1; i < NODDI_IC_MAX_POLYNOMIAL_ORDER + 1; i++){
                    tmp[i] = (-exp(-x) + (i - 0.5) * tmp[i-1]) / x;
                }

                result[0] = tmp[0];
                result[1] = -0.5*tmp[0] + 1.5*tmp[1];
                result[2] = 0.375*tmp[0] - 3.75*tmp[1] + 4.375*tmp[2];
                result[3] = -0.3125*tmp[0] + 6.5625*tmp[1] - 19.6875*tmp[2] + 14.4375*tmp[3];
                result[4] = 0.2734375*tmp[0] - 9.84375*tmp[1] + 54.140625*tmp[2] - 93.84375*tmp[3] + 50.2734375*tmp[4];
                result[5] = -(63/256.0)*tmp[0] + (3465/256.0)*tmp[1] - (30030/256.0)*tmp[2] + (90090/256.0)*tmp[3] - (109395/256.0)*tmp[4] + (46189/256.0)*tmp[5];
                result[6] = (231/1024.0)*tmp[0] - (18018/1024.0)*tmp[1] + (225225/1024.0)*tmp[2] - (1021020/1024.0)*tmp[3] + (2078505/1024.0)*tmp[4] - (1939938/1024.0)*tmp[5] + (676039/1024.0)*tmp[6];
            }
            else{
                // approximate
                mot_float_type tmp[NODDI_IC_MAX_POLYNOMIAL_ORDER - 1];
                tmp[0] = x * x;
                tmp[1] = tmp[0] * x;
                tmp[2] = tmp[1] * x;
                tmp[3] = tmp[2] * x;
                tmp[4] = tmp[3] * x;

                result[0] = 2 - 2*x/3.0 + tmp[0]/5 - tmp[1]/21.0 + tmp[2]/108.0;
                result[1] = -4*x/15.0 + 4*tmp[0]/35.0 - 2*tmp[1]/63.0 + 2*tmp[2]/297.0;
                result[2] = 8*tmp[0]/315.0 - 8*tmp[1]/693.0 + 4*tmp[2]/1287.0;
                result[3] = -16*tmp[1]/9009.0 + 16*tmp[2]/19305.0;
                result[4] = 32*tmp[2]/328185.0;
                result[5] = -64*tmp[3]/14549535.0;
                result[6] = 128*tmp[4]/760543875.0;
            }
        }

        /**
            function [C, D] = WatsonSHCoeff(k)
            Computes the spherical harmonic (SH) coefficients of the Watson's
            distribution with the concentration parameter k (kappa) up to the 12th order.

            Truncating at the 12th order gives good approximation for kappa up to 64.

            Note that the SH coefficients of the odd orders are always zero.

            author: Gary Hui Zhang (gary.zhang@ucl.ac.uk)
        */
        void NODDI_IC_WatsonSHCoeff(const mot_float_type kappa, mot_float_type* const result){
            result[0] = M_SQRTPI_F * 2;

            if(kappa <= 30){
                mot_float_type ks[NODDI_IC_MAX_POLYNOMIAL_ORDER - 1];
                ks[0] = kappa * kappa;
                ks[1] = ks[0] * kappa;
                ks[2] = ks[1] * kappa;
                ks[3] = ks[2] * kappa;
                ks[4] = ks[3] * kappa;

                if(kappa > 0.1){
                    // exact
                    mot_float_type sks[NODDI_IC_MAX_POLYNOMIAL_ORDER];
                    sks[0] = sqrt(kappa);
                    sks[1] = sks[0] * kappa;
                    sks[2] = sks[1] * kappa;
                    sks[3] = sks[2] * kappa;
                    sks[4] = sks[3] * kappa;
                    sks[5] = sks[4] * kappa;

                    mot_float_type erfik = erfi(sks[0]);
                    mot_float_type ierfik = 1/erfik;
                    mot_float_type ek = exp(kappa);
                    mot_float_type dawsonk = M_SQRTPI_2_F * erfik/ek;

                    result[1] = 3 * sks[0] - (3 + 2 * kappa) * dawsonk;
                    result[1] = sqrt(5.0) * result[1] * ek;
                    result[1] = result[1]*ierfik/kappa;

                    result[2] = (105 + 60*kappa + 12*ks[0] )*dawsonk;
                    result[2] = result[2] -105*sks[0] + 10*sks[1];
                    result[2] = .375*result[2]*ek/ks[0];
                    result[2] = result[2]*ierfik;

                    result[3] = -3465 - 1890*kappa - 420*ks[0]  - 40*ks[1] ;
                    result[3] = result[3]*dawsonk;
                    result[3] = result[3] + 3465*sks[0] - 420*sks[1]  + 84*sks[2];
                    result[3] = result[3]*sqrt(13*M_PI_F)/64/ks[1];
                    result[3] = result[3]/dawsonk;

                    result[4] = 675675 + 360360*kappa + 83160*ks[0]  + 10080*ks[1]  + 560*ks[2] ;
                    result[4] = result[4]*dawsonk;
                    result[4] = result[4] - 675675*sks[0] + 90090*sks[1]  - 23100*sks[2]  + 744*sks[3];
                    result[4] = sqrt(17.0)*result[4]*ek;
                    result[4] = result[4]/512.0/ks[2];
                    result[4] = result[4]*ierfik;

                    result[5] = -43648605 - 22972950*kappa - 5405400*ks[0]  - 720720*ks[1]  - 55440*ks[2]  - 2016*ks[3];
                    result[5] = result[5]*dawsonk;
                    result[5] = result[5] + 43648605*sks[0] - 6126120*sks[1]  + 1729728*sks[2]  - 82368*sks[3]  + 5104*sks[4];
                    result[5] = sqrt(21*M_PI_F)*result[5]/4096.0/ks[3];
                    result[5] = result[5]/dawsonk;

                    result[6] = 7027425405 + 3666482820*kappa + 872972100*ks[0]  + 122522400*ks[1]   + 10810800*ks[2]  + 576576*ks[3]  + 14784*ks[4];
                    result[6] = result[6]*dawsonk;
                    result[6] = result[6] - 7027425405*sks[0] + 1018467450*sks[1]  - 302630328*sks[2]  + 17153136*sks[3]  - 1553552*sks[4]  + 25376*sks[5];
                    result[6] = 5*result[6]*ek;
                    result[6] = result[6]/16384.0/ks[4];
                    result[6] = result[6]*ierfik;
                }
                else{
                    // approximate
                    result[1] = (4/3.0*kappa + 8/63.0*ks[0]) * sqrt(M_PI_F/5.0);
                    result[2] = (8/21.0*ks[0] + 32/693.0*ks[1]) * (sqrt(M_PI_F)*0.2);
                    result[3] = (16/693.0*ks[1] + 32/10395.0*ks[2]) * sqrt(M_PI_F/13);
                    result[4] = (32/19305.0*ks[2]) * sqrt(M_PI_F/17);
                    result[5] = 64*sqrt(M_PI_F/21)*ks[3]/692835.0;
                    result[6] = 128*sqrt(M_PI_F)*ks[4]/152108775.0;
                }
            }
            else{
                // large
                mot_float_type lnkd[NODDI_IC_MAX_POLYNOMIAL_ORDER];
                lnkd[0] = log(kappa) - log(30.0);
                lnkd[1] = lnkd[0] * lnkd[0];
                lnkd[2] = lnkd[1] * lnkd[0];
                lnkd[3] = lnkd[2] * lnkd[0];
                lnkd[4] = lnkd[3] * lnkd[0];
                lnkd[5] = lnkd[4] * lnkd[0];

                result[1] = 7.52308 + 0.411538*lnkd[0] - 0.214588*lnkd[1] + 0.0784091*lnkd[2] - 0.023981*lnkd[3] + 0.00731537*lnkd[4] - 0.0026467*lnkd[5];
                result[2] = 8.93718 + 1.62147*lnkd[0] - 0.733421*lnkd[1] + 0.191568*lnkd[2] - 0.0202906*lnkd[3] - 0.00779095*lnkd[4] + 0.00574847*lnkd[5];
                result[3] = 8.87905 + 3.35689*lnkd[0] - 1.15935*lnkd[1] + 0.0673053*lnkd[2] + 0.121857*lnkd[3] - 0.066642*lnkd[4] + 0.0180215*lnkd[5];
                result[4] = 7.84352 + 5.03178*lnkd[0] - 1.0193*lnkd[1] - 0.426362*lnkd[2] + 0.328816*lnkd[3] - 0.0688176*lnkd[4] - 0.0229398*lnkd[5];
                result[5] = 6.30113 + 6.09914*lnkd[0] - 0.16088*lnkd[1] - 1.05578*lnkd[2] + 0.338069*lnkd[3] + 0.0937157*lnkd[4] - 0.106935*lnkd[5];
                result[6] = 4.65678 + 6.30069*lnkd[0] + 1.13754*lnkd[1] - 1.38393*lnkd[2] - 0.0134758*lnkd[3] + 0.331686*lnkd[4] - 0.105954*lnkd[5];
            }
        }
    '''



import os
from collections import OrderedDict

from PyQt5 import QtCore
from PyQt5.QtCore import pyqtSlot, Qt, QObject
from PyQt5.QtGui import QBrush
from PyQt5.QtWidgets import QFileDialog, QTableWidgetItem, QAbstractItemView, QMenu, QMessageBox, \
    QDialog, QDialogButtonBox
from mdt.gui.model_fit.design.ui_generate_protocol_load_gb_dialog import Ui_LoadGBDialog
from mdt.gui.model_fit.design.ui_generate_protocol_update_dialog import Ui_UpdateColumnDialog

import mdt
from mdt.gui.model_fit.design.ui_generate_protocol_tab import Ui_GenerateProtocolTabContent
from mdt.gui.utils import protocol_files_filters, MainTab
from mdt.protocols import Protocol, load_bvec_bval

__author__ = 'Robbert Harms'
__date__ = "2016-06-27"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class GenerateProtocolTab(MainTab, Ui_GenerateProtocolTabContent, QObject):

    def __init__(self, shared_state, computations_thread):
        super().__init__()
        self._shared_state = shared_state
        self._protocol = Protocol()
        self._opened_file = self._shared_state.base_dir
        self._tab_content = None

        self._system_columns = OrderedDict([['#', self._create_volume_number_column]])

    def setupUi(self, tab_content):
        super().setupUi(tab_content)
        self._tab_content = tab_content

        self.loadProtocolButton.clicked.connect(lambda: self._select_protocol())
        self.saveButton.clicked.connect(lambda: self._save_protocol())
        self.loadColumnButton.clicked.connect(lambda: self._load_column_action())
        self.loadGB.clicked.connect(lambda: self._load_g_and_b())
        self.clearButton.clicked.connect(self._clear_table)

        self.protocol_table.setSortingEnabled(True)

        headers = self.protocol_table.horizontalHeader()
        headers.setContextMenuPolicy(Qt.CustomContextMenu)
        headers.customContextMenuRequested.connect(self.show_header_context_menu)
        headers.setSelectionMode(QAbstractItemView.SingleSelection)

    def _select_protocol(self):
        open_file, used_filter = QFileDialog().getOpenFileName(
            caption='Select the protocol', directory=self._shared_state.base_dir,
            filter=';;'.join(protocol_files_filters))

        if open_file:
            self._shared_state.base_dir = os.path.dirname(open_file)
            self.load_protocol(open_file)

    def _save_protocol(self):
        output_file_name, used_filter = QFileDialog().getSaveFileName(
            caption='Save the protocol as', directory=self._opened_file,
            filter=';;'.join(protocol_files_filters))

        if not output_file_name.endswith('.prtcl'):
            output_file_name = output_file_name + '.prtcl'

        if os.path.isdir(os.path.dirname(output_file_name)) and self._protocol.length:
            mdt.write_protocol(self._protocol, output_file_name)
            print('Saved protocol as: {}'.format(output_file_name))

    def _clear_table(self):
        self._protocol = Protocol()
        self._update_views()

    def load_protocol(self, file_name):
        self._protocol = mdt.protocols.load_protocol(file_name)
        self._update_views()
        self._opened_file = file_name
        print('Loaded protocol: {}'.format(file_name))

    def _update_views(self):
        self._update_protocol_info()
        self._update_table_view()

    def _update_protocol_info(self):
        self.nmrRows.setText(str(self._protocol.length))

        try:
            self.nmrUnweighted.setText(str(len(self._protocol.get_unweighted_indices())))
        except KeyError:
            self.nmrUnweighted.setText('0')

        try:
            self.nmrWeighted.setText(str(len(self._protocol.get_weighted_indices())))
        except KeyError:
            self.nmrWeighted.setText('0')

        try:
            shells = self._protocol.get_b_values_shells()
            self.nmrShells.setText(str(len(shells)))
            shells_text = []
            for shell in shells:
                shells_text.append('{0:0=.3f}e9 s/m^2 ({1} volumes)'.format(shell['b_value'] / 1e9,
                                                                            shell['nmr_volumes']))
            self.differentShells.setText(', '.join(shells_text))
        except KeyError:
            self.nmrShells.setText('0')
            self.differentShells.setText('-')

        self.nmrColumns.setText(str(self._protocol.number_of_columns))

    def _update_table_view(self):
        all_column_names, real_column_names, estimated_column_names, system_column_names = self._get_column_names()

        self.protocol_table.clear()
        self.protocol_table.setRowCount(self._protocol.length)
        self.protocol_table.setColumnCount(len(all_column_names))

        for index, column_name in enumerate(all_column_names):
            header_cell = QTableWidgetItem(column_name)
            if column_name in estimated_column_names:
                header_cell.setToolTip('This column is estimated from the other columns in the protocol.')
            self.protocol_table.setHorizontalHeaderItem(index, header_cell)

        for column_ind, column_name in enumerate(all_column_names):
            if column_name in system_column_names:
                generate_function = self._system_columns[column_name]
                cells = generate_function()
                for row, cell in enumerate(cells):
                    self.protocol_table.setItem(row, column_ind, cell)
            else:
                try:
                    values = self._protocol.get_column(column_name)
                    for row in range(self._protocol.length):
                        cell = NumericalSortedTableItem('{:e}'.format(values[row, 0]))
                        cell.setFlags(QtCore.Qt.ItemIsEnabled)

                        if column_name in estimated_column_names:
                            cell.setBackground(QBrush(Qt.lightGray))

                        self.protocol_table.setItem(row, column_ind, cell)
                except KeyError:
                    for row in range(self._protocol.length):
                        cell = QTableWidgetItem('?')
                        cell.setFlags(QtCore.Qt.ItemIsEnabled)
                        cell.setBackground(QBrush(Qt.lightGray))
                        self.protocol_table.setItem(row, column_ind, cell)

        self.protocol_table.resizeColumnsToContents()

    def _get_column_names(self):
        real_column_names = self._protocol.column_names

        if len(real_column_names):
            estimated_column_names = self._protocol.estimated_column_names
        else:
            estimated_column_names = []

        system_column_names = list(self._system_columns.keys())

        all_column_names = system_column_names + real_column_names + estimated_column_names

        return [all_column_names, real_column_names, estimated_column_names, system_column_names]

    def _create_volume_number_column(self):
        """Callback function to generate the volume number column cells.

        This should return a list of cells in the correct order.
        """
        cells = []
        for volume_nmr in range(self._protocol.length):
            cell = NumericalSortedTableItem(str(volume_nmr))
            cell.setFlags(QtCore.Qt.ItemIsEnabled)
            cell.setBackground(QBrush(Qt.lightGray))
            cells.append(cell)
        return cells

    def show_header_context_menu(self, position):
        all_column_names, real_column_names, estimated_column_names, system_column_names = self._get_column_names()
        column_index = self.protocol_table.horizontalHeader().logicalIndexAt(position)
        column_name = all_column_names[column_index]

        if column_name in real_column_names:
            menu = QMenu()
            remove_action = menu.addAction("&Remove column")
            ac = menu.exec_(self.protocol_table.horizontalHeader().mapToGlobal(position))

            if ac == remove_action:
                quit_msg = "Are you sure you want to remove the " \
                           "column '{}' from the protocol".format(column_name)
                reply = QMessageBox.question(self._tab_content, 'Delete confirmation', quit_msg,
                                             QMessageBox.Yes, QMessageBox.No)

                if reply == QMessageBox.Yes:
                    self._protocol = self._protocol.with_column_removed(column_name)
                    self._update_views()

    def _load_column_action(self):
        dialog = LoadColumnDialog(self._shared_state, self._tab_content)
        return_value = dialog.exec_()

        if return_value:
            self._protocol = dialog.update_protocol(self._protocol)
            self._update_views()

    def _load_g_and_b(self):
        dialog = LoadGBDialog(self._shared_state, self._tab_content)
        return_value = dialog.exec_()

        if return_value:
            self._protocol = dialog.get_protocol()
            self._update_views()


class NumericalSortedTableItem(QTableWidgetItem):

    def __lt__(self, other):
        if isinstance(other, QTableWidgetItem):
            try:
                this_value = float(self.text())
                other_value = float(other.text())

                return this_value < other_value
            except ValueError:
                pass

        return super().__lt__(other)


class LoadColumnDialog(Ui_UpdateColumnDialog, QDialog):

    def __init__(self, shared_state, parent):
        super().__init__(parent)
        self._input_options = {'from_file': 0, 'from_value': 1}
        self._shared_state = shared_state
        self.setupUi(self)
        self.inputMethodSelector.currentIndexChanged.connect(self.enable_correct_inputs)
        self.inputMethodSelector.setCurrentIndex(1)
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)
        self.columnNameInput.textChanged.connect(self._update_ok_button)
        self.singleValueInput.textChanged.connect(self._update_ok_button)
        self.fileInput.clicked.connect(lambda: self._select_value_file())
        self.selectedFile.textChanged.connect(self._update_ok_button)
        self.enable_correct_inputs(self.inputMethodSelector.currentIndex())

    def update_protocol(self, protocol):
        column_name = self.columnNameInput.text()
        if column_name:
            try:
                scale = float(self.valueScale.text())
            except ValueError:
                scale = 1

            if self.inputMethodSelector.currentIndex() == self._input_options['from_value']:
                value = float(self.singleValueInput.text())
                return protocol.with_new_column(column_name, value * scale)
            else:
                return protocol.with_added_column_from_file(column_name, self.selectedFile.text(), scale)

    @pyqtSlot(int)
    def enable_correct_inputs(self, selection):
        if selection == self._input_options['from_value']:
            self.singleValueInput.setDisabled(False)
            self.fileInput.setDisabled(True)
            self.selectedFile.setDisabled(True)
        else:
            self.singleValueInput.setDisabled(True)
            self.fileInput.setDisabled(False)
            self.selectedFile.setDisabled(False)

    def _update_ok_button(self):
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(self.columnNameInput.text() != '' and self._has_value())

    def _has_value(self):
        if self.inputMethodSelector.currentIndex() == self._input_options['from_value']:
            if self.singleValueInput.text() != '':
                try:
                    float(self.singleValueInput.text())
                    return True
                except ValueError:
                    pass
            return False
        else:
            if os.path.isfile(self.selectedFile.text()):
                return True
        return False

    def _select_value_file(self):
        initial_dir = self._shared_state.base_dir
        if self.selectedFile.text() != '':
            initial_dir = self.selectedFile.text()

        open_file, used_filter = QFileDialog().getOpenFileName(caption='Select the column info file',
                                                               directory=initial_dir)

        if open_file:
            self.selectedFile.setText(open_file)
            self._shared_state.base_dir = os.path.dirname(open_file)
            self._update_ok_button()


class LoadGBDialog(Ui_LoadGBDialog, QDialog):

    def __init__(self, shared_state, parent):
        super().__init__(parent)
        self._shared_state = shared_state
        self.setupUi(self)
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)
        self.bvecFileInput.textChanged.connect(self._update_ok_button)
        self.bvalFileInput.textChanged.connect(self._update_ok_button)
        self.bvecFileChooser.clicked.connect(lambda: self._select_bvec_file())
        self.bvalFileChooser.clicked.connect(lambda: self._select_bval_file())

    def get_protocol(self):
        try:
            bval_scale = float(self.bvalScale.text())
        except:
            bval_scale = 1

        return load_bvec_bval(bvec=self.bvecFileInput.text(),
                              bval=self.bvalFileInput.text(),
                              bval_scale=bval_scale)

    def _update_ok_button(self):
        enable = os.path.isfile(self.bvalFileInput.text()) and os.path.isfile(self.bvecFileInput.text())
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(enable)

    def _select_bvec_file(self):
        initial_dir = self._shared_state.base_dir
        if self.bvecFileInput.text() != '':
            initial_dir = self.bvecFileInput.text()

        open_file, used_filter = QFileDialog().getOpenFileName(caption='Select the bvec file', directory=initial_dir)

        if open_file:
            self.bvecFileInput.setText(open_file)
            self._shared_state.base_dir = os.path.dirname(open_file)
            self._update_ok_button()

    def _select_bval_file(self):
        initial_dir = self._shared_state.base_dir
        if self.bvalFileInput.text() != '':
            initial_dir = self.bvalFileInput.text()

        open_file, used_filter = QFileDialog().getOpenFileName(caption='Select the bval file', directory=initial_dir)

        if open_file:
            self.bvalFileInput.setText(open_file)
            self._shared_state.base_dir = os.path.dirname(open_file)
            self._update_ok_button()
